"""
Module containing geometry block set operations
"""
import geopandas as gpd
from shapely.geometry import box

from .base import GeometryBlock, BaseSingle

__all__ = ["Difference", "Intersection"]


class Difference(BaseSingle):
    """
    Calculate the geometric difference of two GeometryBlocks.

    All geometries in the source GeometryBlock will be adapted by geometries
    with the same index from the second GeometryBlock. The difference operation
    removes any overlap between the geometries from the first geometry.

    Args:
      source (GeometryBlock): First geometry source.
      other (GeometryBlock): Second geometry source.

    Returns:
      A GeometryBlock with altered geometries. Properties are preserved.
    """

    def __init__(self, source, other):
        if not isinstance(other, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(other)))
        super().__init__(source, other)  # type of source is checked in super()

    @property
    def other(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        if request["mode"] == "extent":
            # the resulting extent might be smaller than the source's extent,
            # but we would have to execute the Difference in order to know.
            return [(self.source, request)]

        # we need to get the extent of the geometries in source, in order to
        # fetch all needed geometries in other
        extent_request = request.copy()
        extent_request["mode"] = "extent"
        extent = self.source.get_data(**extent_request)["extent"]

        if extent is None:  # shortcut when there are no geometries present
            projection = request["projection"]
            return [({"empty": True, "projection": projection}, None)]

        other_request = request.copy()
        other_request["geometry"] = box(*extent)
        return [(self.source, request), (self.other, other_request)]

    @staticmethod
    def process(source_data, other_data=None):
        if other_data is None:
            if source_data.get("empty"):  # catches the no-geometry shortcut
                return {
                    "features": gpd.GeoDataFrame([]),
                    "projection": source_data["projection"],
                }
            else:
                return source_data

        a = source_data["features"]
        b = other_data["features"]
        if len(a) == 0 or len(b) == 0:
            return source_data  # do nothing

        a_series = a["geometry"]
        b_series = b["geometry"].reindex(a_series.index)
        result_series = a_series.difference(b_series)

        # re-insert geometries that were missing in b (we want A - None = A)
        missing_in_b = b_series.isna()
        result_series[missing_in_b] = a_series[missing_in_b]
        result = a.set_geometry(result_series)
        return {"features": result, "projection": source_data["projection"]}


class Intersection(BaseSingle):
    """Calculate the intersection of a GeometryBlock with the request geometry.

    Normally, geometries returned by a GeometryBlock may be partially outside
    of the requested geometry. This block ensures that the geometries are
    strictly inside the requested geometry by taking the intersection of
    each geometry with the request geometry.

    Args:
      source (GeometryBlock): Input geometry source.

    Returns:
      A GeometryBlock with altered geometries. Properties are preserved.
    """

    def __init__(self, source, other=None):
        if isinstance(other, GeometryBlock):
            raise NotImplementedError(
                "Cannot compare geometries with another geometry datasource"
            )
        elif other is not None:
            raise NotImplementedError(
                "Cannot compare geometries with a constant geometry"
            )
        super().__init__(source, other)

    @property
    def other(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        return [(self.source, request), (request["geometry"], None)]

    @staticmethod
    def process(data, geometry):
        # the features will always be in the projection of geometry so we can
        # safely not think about projections here.
        # This cover 'intersects' and 'centroid' mode.
        if "features" in data:
            features = data["features"]
            features = features.set_geometry(features.geometry.intersection(geometry))
            return {"features": features, "projection": data["projection"]}
        elif "extent" in data:
            # intersect the bboxes
            bbox1 = data["extent"]
            bbox2 = geometry.bounds
            bbox = (
                max(bbox1[0], bbox2[0]),
                max(bbox1[1], bbox2[1]),
                min(bbox1[2], bbox2[2]),
                min(bbox1[3], bbox2[3]),
            )
            return {"extent": bbox, "projection": data["projection"]}



"""
Module containing reduction raster blocks.
"""
import numpy as np
from dask_geomodeling.utils import filter_none, get_index
from dask_geomodeling.utils import parse_percentile_statistic
from .base import RasterBlock
from .elemwise import BaseElementwise
from functools import partial

__all__ = ["Max"]

STATISTICS = {
    "first": None,
    "last": None,
    "count": None,
    "sum": np.nansum,
    "mean": np.nanmean,
    "min": np.nanmin,
    "max": np.nanmax,
    "argmin": np.nanargmin,
    "argmax": np.nanargmax,
    "std": np.nanstd,
    "var": np.nanvar,
    "median": np.nanmedian,
    "product": np.nanprod,
    # "p<number>" uses np.nanpercentile
}


def check_statistic(statistic):
    if statistic not in STATISTICS:
        percentile = parse_percentile_statistic(statistic)
        if percentile is None:
            raise ValueError('Unknown statistic "{}"'.format(statistic))


def reduce_rasters(stack, statistic, no_data_value=None, dtype=None):
    """Apply a statistic (e.g. "mean") to a stack of rasters, skipping
    'no data' values.

    In this context, reduce means that the dimensionality of the input data
    is reduced by one.

    Args:
      stack (list): a list of dicts containing "values" (ndarray)
        and "no_data_value". If the list has zero length or if the ndarrays
        do not have the same shape, a ValueError is raised.
      statistic (str): the applied statistic (no data is ignored). One of:
        {"last", "first", "count", "sum", "mean", "min",
        "max", "argmin", "argmax", "product", "std", "var", "p<number>"}
      no_data_value (number, optional): the 'no data' value in the output
        array. Defaults to the no data value of the first element in the stack.
      dtype (str or dtype): the datatype of the output array. Defaults to the
        dtype of the first element in the stack. If the input
        data cannot be cast to this dtype, a ValueError is raised.

    Returns:
      dict with "values" and "no_data_value"
    """
    if statistic not in STATISTICS:
        percentile = parse_percentile_statistic(statistic)
        if percentile is None:
            raise KeyError('Unknown statistic "{}"'.format(statistic))
        else:
            statistic = "percentile"

    if len(stack) == 0:
        raise ValueError("Cannot reduce a zero-length stack")

    # get the output array properties (dtype, no_data_value, shape)
    if dtype is None:
        dtype = stack[0]["values"].dtype
    if no_data_value is None:
        no_data_value = stack[0]["no_data_value"]
    shape = stack[0]["values"].shape

    # sum, count and nans output do not contain no data: fill zeroes right away
    if statistic in {"sum", "count", "nans"}:
        fill_value = 0
    else:
        fill_value = no_data_value

    # create the output array
    out = np.full(shape, fill_value, dtype)

    if statistic == "last":
        # populate 'out' with the last value that is not 'no data'
        for data in stack:
            index = get_index(data["values"], data["no_data_value"])
            out[index] = data["values"][index]
    elif statistic == "first":
        # populate 'out' with the first value that is not 'no data'
        for data in stack[::-1]:
            index = get_index(data["values"], data["no_data_value"])
            out[index] = data["values"][index]
    elif statistic == "count":
        # count the number of values that are not 'no data'
        for data in stack:
            out += get_index(data["values"], data["no_data_value"])
    else:
        if statistic == "percentile":
            func = partial(np.nanpercentile, q=percentile)
        else:
            func = STATISTICS[statistic]
        # transform 'no data' into 'nan' to be able to use numpy functions
        # NB: the dtype is at least float16 to accomodate NaN
        stack_array = np.full(
            (len(stack),) + shape, np.nan, np.result_type(dtype, np.float16)
        )
        for i, data in enumerate(stack):
            index = get_index(data["values"], data["no_data_value"])
            stack_array[i, index] = data["values"][index]

        # protect against all-NaN slice warnings and errors
        not_all_nan = ~np.all(np.isnan(stack_array), axis=0)

        # perform the math
        out[not_all_nan] = func(stack_array[:, not_all_nan], axis=0)

    return {"values": out, "no_data_value": no_data_value}


class BaseReduction(BaseElementwise):
    def __init__(self, *args):
        for arg in args:
            if not isinstance(arg, RasterBlock):
                raise TypeError("'{}' object is not allowed".format(type(arg)))
        super().__init__(*args)

    def get_sources_and_requests(self, **request):
        """
        This method is more strict than the one from BaseElementWise, it is not
        possible to get data from this block when its period is None, meaning
        that either the sources have no common period, or one of the sources'
        periods is None.
        """
        period = self.period
        process_kwargs = {
            "dtype": self.dtype.name, "fillvalue": self.fillvalue,
        }
        if period is None:
            return [(process_kwargs, None)]

        # limit request to self.period so that resulting data is aligned
        start = request.get("start", None)
        stop = request.get("stop", None)
        if start is not None:
            if stop is not None:
                request["start"] = max(start, period[0])
                request["stop"] = min(stop, period[1])
            else:  # stop is None but start isn't
                request["start"] = min(max(start, period[0]), period[1])
        else:  # start and stop are both None
            request["start"] = period[1]

        sources_and_requests = [(source, request) for source in self.args]

        return [(process_kwargs, None)] + sources_and_requests

    @property
    def extent(self):
        """ Boundingbox of combined contents in WGS84 projection. """
        extents = filter_none([x.extent for x in self.args])
        if len(extents) == 0:
            return None
        elif len(extents) == 1:
            return extents[0]

        # multiple extents: return the joined box
        x1 = min([e[0] for e in extents])
        y1 = min([e[1] for e in extents])
        x2 = max([e[2] for e in extents])
        y2 = max([e[3] for e in extents])
        return x1, y1, x2, y2

    @property
    def geometry(self):
        """Combined geometry in the projection of the first store geometry. """
        geometries = filter_none([x.geometry for x in self.args])
        if len(geometries) == 0:
            return
        elif len(geometries) == 1:
            return geometries[0]
        result = geometries[0]
        sr = result.GetSpatialReference()
        for geometry in geometries[1:]:
            if not geometry.GetSpatialReference().IsSame(sr):
                geometry = geometry.Clone()
                geometry.TransformTo(sr)
            result = result.Union(geometry)
        return result


def wrap_reduction_function(statistic):
    def reduction_function(process_kwargs, *args):
        stack = []
        for arg in args:
            if "time" in arg or "meta" in arg:
                # return the time / meta right away. assumes there are no
                # mixed requests and that time is aligned
                return arg
            if arg is None:
                continue
            stack.append(arg)

        # return None if all source data is None
        if len(stack) == 0:
            return

        # see BaseElementWise.get_sources_and_requests
        return reduce_rasters(
            stack,
            statistic,
            process_kwargs["fillvalue"],
            process_kwargs["dtype"],
        )
    return reduction_function


class Max(BaseReduction):
    """
    Take the maximum value of two or more rasters, ignoring no data.

    Args:
      *args (list of RasterBlocks): list of rasters to be combined.

    Returns:
      RasterBlock with the maximum values
    """
    process = staticmethod(wrap_reduction_function("max"))

    @property
    def dtype(self):
        # skip the default behaviour where we use at least int32 / float32
        return np.result_type(*self.args)



"""
Module containing raster blocks that combine rasters.
"""
import itertools
from datetime import timedelta as Timedelta
import numpy as np

from dask_geomodeling.utils import filter_none, get_dtype_max, get_index
from dask_geomodeling.utils import GeoTransform

from .base import RasterBlock

__all__ = ["Group"]


class BaseCombine(RasterBlock):
    """ Base block that combines rasters into a larger one.

    The ancestor stores are kept in ``self.args``. Attributes are greedy:
    ``period`` is the union of the ancestor periods, and ``extent`` the union
    of the ancestor extents. The ``timedelta`` is propagated only if the
    ancestor stores have equal ``timedelta`` and if they are aligned. Rasters
    without data are ignored.
    """

    def __init__(self, *args):
        for arg in args:
            if not isinstance(arg, RasterBlock):
                raise TypeError("'{}' object is not allowed".format(type(arg)))
        super(BaseCombine, self).__init__(*args)

    @staticmethod
    def get_aligned_timedelta(sources):
        """ Checks provided sources and returns the timedelta when all sources
        are aligned. Stores without data are ignored. """
        timedeltas = []
        periods = []
        for arg in sources:
            timedelta, period = arg.timedelta, arg.period
            if period is not None and timedelta is not None:
                timedeltas.append(timedelta)
                periods.append(period)

        if len(timedeltas) == 0:
            return None
        elif len(timedeltas) == 1:
            return timedeltas[0]

        # multiple timedeltas: return None if not equal
        if not timedeltas[1:] == timedeltas[:-1]:
            return None
        else:
            # the periods must be spaced an integer times timedelta apart
            timedelta_sec = timedeltas[0].total_seconds()
            first, _ = periods[0]
            for a, _ in periods[1:]:
                if (first - a).total_seconds() % timedelta_sec != 0:
                    return None
            return timedeltas[0]

    @property
    def timedelta(self):
        """ The period between timesteps in case of equidistant time. """
        return self.get_aligned_timedelta(self.args)

    @property
    def period(self):
        """ Return the combined period datetime tuple. """
        periods = filter_none([x.period for x in self.args])
        if len(periods) == 0:
            return None
        elif len(periods) == 1:
            return periods[0]

        # multiple periods: return the joined period
        return min([p[0] for p in periods]), max([p[1] for p in periods])

    @property
    def extent(self):
        """ Boundingbox of combined contents in WGS84 projection. """
        extents = filter_none([x.extent for x in self.args])
        if len(extents) == 0:
            return None
        elif len(extents) == 1:
            return extents[0]

        # multiple extents: return the joined box
        x1 = min([e[0] for e in extents])
        y1 = min([e[1] for e in extents])
        x2 = max([e[2] for e in extents])
        y2 = max([e[3] for e in extents])
        return x1, y1, x2, y2

    @property
    def dtype(self):
        return np.result_type(*self.args)

    @property
    def fillvalue(self):
        return get_dtype_max(self.dtype)

    @property
    def geometry(self):
        """Combined geometry in the projection of the first store geometry. """
        geometries = filter_none([x.geometry for x in self.args])
        if len(geometries) == 0:
            return
        elif len(geometries) == 1:
            return geometries[0]
        result = geometries[0]
        sr = result.GetSpatialReference()
        for geometry in geometries[1:]:
            if not geometry.GetSpatialReference().IsSame(sr):
                geometry = geometry.Clone()
                geometry.TransformTo(sr)
            result = result.Union(geometry)
        return result

    @property
    def projection(self):
        """Projection of the data if they match, else None"""
        projection = self.args[0].projection
        if projection is None:
            return
        for arg in self.args[1:]:
            if projection != arg.projection:
                return
        return projection

    @property
    def geo_transform(self):
        geo_transform = self.args[0].geo_transform
        if geo_transform is None:
            return
        geo_transform = GeoTransform(geo_transform)
        for arg in self.args[1:]:
            other = arg.geo_transform
            if other is None or not geo_transform.aligns_with(other):
                return
        return geo_transform


class Group(BaseCombine):
    """
    Combine multiple rasters into a single one.

    Operation to combine multiple rasters into one along all three axes (x, y
    and temporal). To only fill 'no data' values of input rasters that have the
    same temporal resolution ``dask_geomodeling.raster.elemwise.FillNoData``
    is preferred.

    Values at equal timesteps in the contributing rasters are considered
    starting with the leftmost input raster. Therefore, values from rasters
    that are more 'to the right' are shown in the result. 'no data' values are
    transparent and will show data of rasters more 'to the left'.

    Args:
      *args (list of RasterBlocks): list of rasters to be combined.

    Returns:
      RasterBlock that combines all input rasters
    """

    def get_stores(self, start, stop):
        """ Return all relevant stores for given start and stop. """
        # check stores and select those stores that contain data
        stores = [s for s in self.args if s.period is not None]

        # there could be no store with data at all
        if not stores:
            return stores

        # convenience
        starts, stops = zip(*(s.period for s in stores))

        # pick latest store(s) in time
        if start is None:
            last = max(stops)
            return [s for b, s in zip(stops, stores) if b == last]

        if stop is None:
            # return any stores that contain start
            zipped = zip(starts, stops, stores)
            result = [s for a, b, s in zipped if a <= start <= b]
            if result:
                return result

            # no store contained start, return closest stores
            closest = min(starts + stops, key=lambda d: abs(d - start))
            zipped = zip(stops + starts, stores + stores)
            return [s for d, s in zipped if d == closest]

        # start and stop given, return all relevant stores
        zipped = zip(starts, stops, stores)
        return [s for a, b, s in zipped if not (stop < a or start > b)]

    def get_sources_and_requests(self, **request):
        start = request.get("start", None)
        stop = request.get("stop", None)
        mode = request["mode"]

        sources = self.get_stores(start, stop)

        # just pass on the source and request if we only have one (or none)
        if len(sources) <= 1:
            requests = [(s, request) for s in sources]
            return [(dict(combine_mode="simple"), None)] + requests

        # plan for merging
        timedelta = self.get_aligned_timedelta(sources)
        mixed_time = timedelta is None or start is None or stop is None

        if mixed_time:  # merge by time
            requests = []
            time_requests = []
            for source in sources:
                # add the stores and requests:
                requests.append((source, request))

                # in case we need the time information, add time requests
                if mode != "time":
                    time_request = dict(mode="time", start=start, stop=stop)
                    time_requests.append((source, time_request))

            process_kwargs = dict(
                combine_mode="by_time", mode=mode, start=start, stop=stop
            )

            # note that time_requests is empty if mode is 'time'
            requests = requests + time_requests
        else:  # merge by bands
            td_sec = timedelta.total_seconds()
            period = self.period
            origin = sources[0].period[0]  # any will do; they are aligned

            if start <= period[0]:
                start = period[0]
            else:
                # ceil start to the closest integer timedelta
                start_delta = (origin - start).total_seconds() % td_sec
                start += Timedelta(seconds=start_delta)

            if stop >= period[1]:
                stop = period[1]
            else:
                # floor stop to the closest integer timedelta
                stop_delta = (stop - origin).total_seconds() % td_sec
                stop -= Timedelta(seconds=stop_delta)

            if mode == "time":
                return [
                    (
                        dict(
                            combine_mode="by_bands",
                            mode=mode,
                            start=start,
                            stop=stop,
                            timedelta=timedelta,
                        ),
                        None,
                    )
                ]

            requests = []
            bands = []
            for source in sources:
                # compute 'bands': the index ranges into the result array
                this_start = max(start, source.period[0])
                this_stop = min(stop, source.period[1])
                first_i = int((this_start - start).total_seconds() // td_sec)
                last_i = int((this_stop - start).total_seconds() // td_sec)
                bands.append((first_i, last_i + 1))

                this_request = request.copy()
                this_request.update(start=this_start, stop=this_stop)
                requests.append((source, this_request))

            process_kwargs = dict(combine_mode="by_bands", mode=mode, bands=bands)

        # in case of a 'vals' request, keep track of the dtype
        if mode == "vals":
            process_kwargs["dtype"] = self.dtype

        return [(process_kwargs, None)] + requests

    @staticmethod
    def _unique_times(multi):
        times = filter_none([data.get("time", None) for data in multi])
        return sorted(set(itertools.chain(*times)))

    @staticmethod
    def _nearest_index(time, start):
        if start is None:
            # last band
            return len(time) - 1
        else:
            # nearest band
            return min(enumerate(time), key=lambda d: abs(d[1] - start))[0]

    @staticmethod
    def _merge_vals_by_time(multi, times, kwargs):
        """ Merge chunks using indices. """
        # determine the unique times and assign result bands
        sorted_times = Group._unique_times(times)
        bands = dict((y, x) for x, y in enumerate(sorted_times))
        fillvalue = get_dtype_max(kwargs["dtype"])

        # initialize values array
        shape = (len(sorted_times),) + multi[0]["values"].shape[1:]
        values = np.full(shape, fillvalue, dtype=kwargs["dtype"])

        # populate values array
        for data, time in zip(multi, times):
            # source_index is the index into the source array
            for source_index, datetime in enumerate(time["time"]):
                source_band = data["values"][source_index]
                # determine data index
                index = get_index(
                    values=source_band, no_data_value=data["no_data_value"]
                )
                # find corresponding target band by source datetime
                target_band = values[bands[datetime]]
                # paste source into target provided there is data
                target_band[index] = source_band[index]

        # check if single band result required
        start, stop = kwargs["start"], kwargs["stop"]
        if stop is None and len(sorted_times) > 1:
            index = Group._nearest_index(sorted_times, start)
            values = values[index : index + 1]

        return {"values": values, "no_data_value": fillvalue}

    @staticmethod
    def _merge_meta_by_time(multi, times, kwargs):
        # determine the unique times and assign result bands
        sorted_times = Group._unique_times(times)
        bands = dict((y, x) for x, y in enumerate(sorted_times))
        meta_result = [None] * len(sorted_times)

        # populate result array
        for data, time in zip(multi, times):
            # source_index is the index into the source array
            for source_index, datetime in enumerate(time["time"]):
                source_band = data["meta"][source_index]
                # find corresponding target band by source datetime
                target_band = bands[datetime]
                # paste source into target provided there is data
                meta_result[target_band] = source_band

        # check if single band result required
        start, stop = kwargs["start"], kwargs["stop"]
        if stop is None and len(sorted_times) > 1:
            index = Group._nearest_index(sorted_times, start)
            meta_result = meta_result[index : index + 1]

        return {"meta": meta_result}

    @staticmethod
    def _merge_vals_by_bands(multi, bands, dtype):
        """ Merge chunks using slices. """
        # analyze band structure
        starts, stops = zip(*bands)
        fillvalue = get_dtype_max(dtype)

        # initialize values array
        shape = (max(stops),) + multi[0]["values"].shape[1:]
        values = np.full(shape, fillvalue, dtype=dtype)

        # populate values array
        for data, (a, b) in zip(multi, bands):
            # index is where the source has data
            index = get_index(
                values=data["values"], no_data_value=data["no_data_value"]
            )
            values[a:b][index] = data["values"][index]

        return {"values": values, "no_data_value": fillvalue}

    @staticmethod
    def _merge_meta_by_bands(multi, bands):
        """ Merge metadata by bands. """
        # analyze band structure
        starts, stops = zip(*bands)

        # initialize metadata list
        meta_result = [""] * max(stops)

        # populate metadata list
        for data, (a, b) in zip(multi, bands):
            for i, meta in zip(range(a, b), data["meta"]):
                if meta:
                    meta_result[i] = meta

        return {"meta": meta_result}

    @staticmethod
    def process(process_kwargs, *args):
        # plan for merging
        combine_mode = process_kwargs["combine_mode"]
        mode = process_kwargs.get("mode", None)

        if combine_mode == "simple":
            # simple mode, return the result right away
            if len(args) == 0:
                return None
            else:
                return args[0]
        elif combine_mode == "by_time" and mode == "time":
            sorted_times = Group._unique_times(args)

            # check if single band result required
            start, stop = process_kwargs["start"], process_kwargs["stop"]
            if stop is None and len(sorted_times) > 1:
                index = Group._nearest_index(sorted_times, start)
                sorted_times = sorted_times[index : index + 1]
            return {"time": sorted_times}
        elif combine_mode == "by_time" and mode in ["meta", "vals"]:
            # split the data and time results, skipping None
            n = int(len(args) // 2)

            # assume that we have None at the same positions
            multi, times = filter_none(args[:n]), filter_none(args[n:])
            if len(multi) == 0:
                return None

            if mode == "vals":
                return Group._merge_vals_by_time(multi, times, process_kwargs)
            elif mode == "meta":
                return Group._merge_meta_by_time(multi, times, process_kwargs)
        elif combine_mode == "by_bands" and mode == "time":
            # start and stop are aligned so we can compute the times here
            start = process_kwargs["start"]
            stop = process_kwargs["stop"]
            delta = process_kwargs["timedelta"]
            length = (stop - start).total_seconds() // delta.total_seconds()
            length = int(length) + 1  # the result includes the last frame
            return {"time": [start + i * delta for i in range(int(length))]}
        elif combine_mode == "by_bands" and mode in ["meta", "vals"]:
            # list the data and bands results, skipping None
            multi = []
            bands = []
            for data, _bands in zip(args, process_kwargs["bands"]):
                if data is None:
                    continue
                multi.append(data)
                bands.append(_bands)

            if len(multi) == 0:
                return None

            if mode == "vals":
                dtype = process_kwargs["dtype"]
                return Group._merge_vals_by_bands(multi, bands, dtype)
            elif mode == "meta":
                return Group._merge_meta_by_bands(multi, bands)
        else:
            raise ValueError("Unknown combine_mode / mode combination")



"""
Module containing text column operations that act on geometry blocks
"""
import numpy as np
import pandas as pd
import re

from .base import GeometryBlock, BaseSingle

__all__ = ["ParseTextColumn"]

# https://catonmat.net/my-favorite-regex
# key matches any ASCII char except for '='
# value matches any ASCII char
REGEX_KEYVALUE = re.compile(r"((?:[ -<>-~])+)=((?:[ -~])*)")


def autocast_value(value):
    """Cast string to string, float, bool or None. """
    if value is None:
        return
    value_lcase = value.lower()
    if value_lcase == "null":
        return
    if value_lcase == "false":
        return False
    if value_lcase == "true":
        return True
    try:
        return float(value)
    except ValueError:
        return value


class ParseTextColumn(BaseSingle):
    """Parses a text column into (possibly multiple) value columns.

    Key, value pairs need to be separated by an equal (``=``) sign.

    Args:
      source (GeometryBlock): Data source
      source_column (str): Existing column in source.
      key_mapping (dict): Mapping containing pairs {key_name: column_name}:

    """

    def __init__(self, source, source_column, key_mapping):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed.".format(type(source)))
        if not isinstance(source_column, str):
            raise TypeError("'{}' object is not allowed.".format(type(source_column)))
        if source_column not in source.columns:
            raise KeyError("Column '{}' is not available.".format(source_column))
        if not isinstance(key_mapping, dict):
            raise TypeError("'{}' object is not allowed.".format(type(key_mapping)))
        super().__init__(source, source_column, key_mapping)

    @property
    def source(self):
        return self.args[0]

    @property
    def source_column(self):
        return self.args[1]

    @property
    def key_mapping(self):
        return self.args[2]

    @property
    def columns(self):
        return self.source.columns | set(self.key_mapping.values())

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "source_column": self.source_column,
            "key_mapping": self.key_mapping,
        }
        return [(self.source, request), (process_kwargs, None)]

    @staticmethod
    def process(data, kwargs):
        source_column = kwargs["source_column"]
        key_mapping = kwargs["key_mapping"]

        if "features" not in data or len(data["features"]) == 0:
            return data  # do nothing for non-feature requests

        f = data["features"].copy()
        # We parse every unique string only a single time by transforming to
        # categorical dtype:
        column = f[source_column].astype("category")

        if len(column.cat.categories) == 0:
            # no data to parse: add empty columns and return directly
            for col in key_mapping.values():
                f[col] = np.nan
            return {"features": f, "projection": data["projection"]}

        def parser(description):
            pairs = dict(REGEX_KEYVALUE.findall(description))
            return [autocast_value(pairs.get(key)) for key in key_mapping.keys()]

        # Parse each category
        extra_columns = pd.DataFrame(
            [parser(x) for x in column.cat.categories], columns=key_mapping.values()
        )

        # Align the generated dataframe with the original. Pandas versions
        # later than 0.19 have a pd.align that could be used also.
        try:
            extra_columns_aligned = extra_columns.reindex(column.cat.codes)
            extra_columns_aligned.index = f.index
        except KeyError:
            extra_columns_aligned = pd.DataFrame([], columns=key_mapping.values())

        # Assign the extra columns to the original dataframe.
        for name in extra_columns_aligned.columns:
            if extra_columns_aligned[name].isnull().all():
                f[name] = np.nan
            else:
                f[name] = extra_columns_aligned[name]

        return {"features": f, "projection": data["projection"]}



from .base import *  # NOQA
from .sources import *  # NOQA
from .aggregate import *  # NOQA
from .set_operations import *  # NOQA
from .constructive import *  # NOQA
from .geom_operations import *  # NOQA
from .field_operations import *  # NOQA
from .parallelize import *  # NOQA
from .merge import *  # NOQA
from .text import *  # NOQA



"""
Module containing raster blocks for spatial operations.
"""
import math

from scipy import ndimage
import numpy as np
from osgeo import ogr

from dask_geomodeling.utils import (
    EPSG3857,
    EPSG4326,
    POLYGON,
    get_sr,
    Extent,
    get_dtype_min,
    get_footprint,
    get_index,
    shapely_transform,
)
from dask_geomodeling.raster.reduction import reduce_rasters, check_statistic

from .base import BaseSingle, RasterBlock
from shapely.geometry import Point

__all__ = ["Dilate", "Smooth", "MovingMax", "HillShade", "Place"]


def expand_request_pixels(request, radius=1):
    """ Expand request by `radius` pixels. Returns None for non-vals requests
    or point requests. """
    if request["mode"] != "vals":  # do nothing with time and meta requests
        return None

    width, height = request["width"], request["height"]
    x1, y1, x2, y2 = request["bbox"]
    pwidth, pheight = x2 - x1, y2 - y1

    if pwidth == 0 or pheight == 0:  # cannot dilate a point request
        return None

    amount_x = pwidth / width * radius
    amount_y = pheight / height * radius

    new_request = request.copy()
    new_request["bbox"] = (x1 - amount_x, y1 - amount_x, x2 + amount_y, y2 + amount_y)
    new_request["width"] += 2 * radius
    new_request["height"] += 2 * radius
    return new_request


def expand_request_meters(request, radius_m=1):
    """
    Expand request by `radius_m` meters, rounded so that an integer number of
    pixels is added to all sides.

    Returns a tuple of:
     - new request with adapted bbox, width and height
     - the radius transformed to pixels as a (y, x) tuple of floats
     - the added margins as a (y, x) tuple of integers
    """
    sr = get_sr(request["projection"])
    bbox = request["bbox"]

    # throughout, variables in the projected unit ( = meters, mostly) are
    # suffixed by _m, in pixels by _px
    if sr.IsGeographic():
        # expand geographic bbox in EPSG3857
        extent_geom = Extent(bbox, sr)
        bbox = extent_geom.transformed(EPSG3857).bbox
    else:
        # most Projected projections are in meters, but to be sure:
        radius_m /= sr.GetLinearUnits()

    # compute the initial zoom factors: how much to expand the bbox to obtain
    # margins of exactly 'radius' (in meters)
    x1, y1, x2, y2 = bbox
    shape_m = y2 - y1, x2 - x1

    if shape_m[0] > 0 and shape_m[1] > 0:
        # Resolution in pixels per meter:
        resolution = request["height"] / shape_m[0], request["width"] / shape_m[1]
        # How many pixels to add:
        radius_px = [radius_m * res for res in resolution]
        # How many pixels to add, rounded to integers:
        margins_px = [int(round(r)) for r in radius_px]
        # How many meters to add (based on rounded pixels):
        margins_m = [m / res for m, res in zip(margins_px, resolution)]
    else:
        # There is no resolution. Add MARGIN_THRESHOLD pixels to the request.
        radius_px = margins_px = [Smooth.MARGIN_THRESHOLD] * 2
        # Expand the request with radius_m exactly.
        margins_m = [radius_m] * 2

    # assemble the request
    new_request = request.copy()
    new_request["bbox"] = (
        x1 - margins_m[1],
        y1 - margins_m[0],
        x2 + margins_m[1],
        y2 + margins_m[0],
    )
    if sr.IsGeographic():
        # transform back to original projection
        extent_proj = Extent(new_request["bbox"], EPSG3857)
        new_request["bbox"] = extent_proj.transformed(sr).bbox
    new_request["height"] += 2 * margins_px[0]
    new_request["width"] += 2 * margins_px[1]

    return new_request, radius_px


class Dilate(BaseSingle):
    """
    Perform spatial dilation on specific cell values.

    Cells with values in the supplied list are spatially dilated by one cell
    in each direction, including diagonals.

    Dilation is performed in the order of the values parameter.

    Args:
      store (RasterBlock): Raster to perform dilation on.
      values (list): Only cells with these values are dilated.

    Returns:
      RasterBlock where cells in values list are dilated.

    See also:
      https://en.wikipedia.org/wiki/Dilation_%28morphology%29
    """

    def __init__(self, store, values):
        values = np.asarray(values, dtype=store.dtype)
        super().__init__(store, values.tolist())

    @property
    def values(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        new_request = expand_request_pixels(request, radius=1)
        if new_request is None:  # not an expandable request: do nothing
            return [(self.store, request)]
        else:
            return [(self.store, new_request), (self.values, None)]

    @staticmethod
    def process(data, values=None):
        if data is None or values is None or "values" not in data:
            return data
        original = data["values"]
        dilated = original.copy()
        for value in np.asarray(values, dtype=original.dtype):
            dilated[ndimage.binary_dilation(original == value)] = value
        dilated = dilated[:, 1:-1, 1:-1]
        return {"values": dilated, "no_data_value": data["no_data_value"]}


class MovingMax(BaseSingle):
    """
    Apply a spatial maximum filter to the data using a circular footprint.

    This can be used for visualization of sparse data.

    Args:
      store (RasterBlock): Raster to which the filter is applied
      size (integer): Diameter of the circular footprint. This should always be
        an odd number larger than 1.

    Returns:
      RasterBlock with maximum values inside the footprint of each input cell.
    """

    def __init__(self, store, size):
        # round size to nearest odd integer
        size = int(2 * round((size - 1) / 2) + 1)
        if size < 3:
            raise ValueError("The size should be odd and larger than 1")
        super(MovingMax, self).__init__(store, size)

    @property
    def size(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        size = self.size
        new_request = expand_request_pixels(request, radius=int(size // 2))
        if new_request is None:  # not an expandable request: do nothing
            return [(self.store, request)]
        else:
            return [(self.store, new_request), (size, None)]

    @staticmethod
    def process(data, size=None):
        if data is None or size is None or "values" not in data:
            return data
        radius = int(size // 2)
        footprint = get_footprint(size)[np.newaxis]

        # put absolute minimum on no data pixels
        array = data["values"].copy()
        minimum = get_dtype_min(array.dtype)
        no_data_mask = array == data["no_data_value"]
        array[no_data_mask] = minimum

        # apply maximum filter
        filtered = ndimage.maximum_filter(array, footprint=footprint)

        # replace absolute minimum with original fillvalue
        filtered[(filtered == minimum) & no_data_mask] = data["no_data_value"]

        # cut out the result
        filtered = filtered[:, radius:-radius, radius:-radius]
        return {"values": filtered, "no_data_value": data["no_data_value"]}


class Smooth(BaseSingle):
    """
    Smooth the values from a raster spatially using Gaussian smoothing.

    Args:
      store (RasterBlock): Raster to be smoothed
      size (number): The extent of the smoothing in meters. The 'sigma' value
        for the Gaussian kernal equals ``size / 3``.
      fill (number): 'no data' are replaced by this value during smoothing,
        defaults to 0.

    Returns:
      RasterBlock with spatially smoothed values.

    See Also:
      https://en.wikipedia.org/wiki/Gaussian_blur

    """

    MARGIN_THRESHOLD = 6

    def __init__(self, store, size, fill=0):
        for x in (size, fill):
            if not isinstance(x, (int, float)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(Smooth, self).__init__(store, size, fill)

    @property
    def size(self):
        return self.args[1]

    @property
    def fill(self):
        return self.args[2]

    def get_sources_and_requests(self, **request):
        if request["mode"] != "vals":  # do nothing with time and meta requests
            return [(self.store, request)]

        new_request, size = expand_request_meters(request, self.size)

        # check how many pixels will be added by the request
        if any([s > self.MARGIN_THRESHOLD for s in size]):
            smooth_mode = "zoom"
            # rescale the size
            zoom = [new_request[x] / request[x] for x in ("height", "width")]
            size = [s / z for s, z in zip(size, zoom)]
            # request the original (not expanded) shape
            new_request["height"] = request["height"]
            new_request["width"] = request["width"]
        else:
            smooth_mode = "exact"

        process_kwargs = dict(smooth_mode=smooth_mode, fill=self.fill, size=size)

        return [(self.store, new_request), (process_kwargs, None)]

    @staticmethod
    def process(data, process_kwargs=None):
        if data is None or process_kwargs is None:
            return data
        smooth_mode = process_kwargs["smooth_mode"]
        size_px = process_kwargs["size"]
        fill = process_kwargs["fill"]

        # fill in nodata values
        values = data["values"].copy()
        no_data_value = data["no_data_value"]
        values[values == no_data_value] = fill

        # compute the sigma
        sigma = 0, size_px[0] / 3, size_px[1] / 3
        ndimage.gaussian_filter(
            values, sigma, output=values, mode="constant", cval=fill
        )

        # remove the margins
        if smooth_mode == "exact":
            my, mx = [int(round(s)) for s in size_px]
            values = values[:, my : values.shape[1] - my, mx : values.shape[2] - mx]
        else:
            _, ny, nx = values.shape
            zy, zx = [1 - 2 * size_px[0] / ny, 1 - 2 * size_px[1] / nx]

            values = ndimage.affine_transform(
                values,
                order=0,
                matrix=np.diag([1, zy, zx]),
                offset=[0, size_px[0], size_px[1]],
            )

        return {"values": values, "no_data_value": no_data_value}


class HillShade(BaseSingle):
    """
    Calculate a hillshade from the raster values.

    Args:
      store (RasterBlock): Raster to which the hillshade algorithm is applied.
      altitude (number): Light source altitude in degrees, defaults to 45.
      azimuth (number): Light source azimuth in degrees, defaults to 315.
      fill (number): Fill value to be used for 'no data' values.

    Returns:
      Hillshaded raster

    See also:
      https://pro.arcgis.com/en/pro-app/tool-reference/3d-analyst/how-hillshade-works.htm
    """

    def __init__(self, store, altitude=45, azimuth=315, fill=0):
        for x in (altitude, azimuth, fill):
            if not isinstance(x, (int, float)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(HillShade, self).__init__(store, float(altitude), float(azimuth), fill)

    @property
    def altitude(self):
        return self.args[1]

    @property
    def azimuth(self):
        return self.args[2]

    @property
    def fill(self):
        return self.args[3]

    @property
    def dtype(self):
        return np.dtype("u1")

    @property
    def fillvalue(self):
        return 256  # on purpose, it does not exist in bytes

    @staticmethod
    def process(data, process_kwargs=None):
        """
        Adapted from:
        https://github.com/OSGeo/gdal/blob/2.0/gdal/apps/gdaldem.cpp#L481

        Edges are not implemented, result clips one pixel from array.
        """
        if process_kwargs is None:
            return data

        array = data["values"].copy()
        array[array == data["no_data_value"]] = process_kwargs["fill"]

        xres, yres = process_kwargs["resolution"]
        alt = math.radians(process_kwargs["altitude"])
        az = math.radians(process_kwargs["azimuth"])
        zsf = 1 / 8  # vertical scale factor
        square_zsf = zsf * zsf

        # gradient
        s0 = slice(None, None), slice(None, -2), slice(None, -2)
        s1 = slice(None, None), slice(None, -2), slice(1, -1)
        s2 = slice(None, None), slice(None, -2), slice(2, None)
        s3 = slice(None, None), slice(1, -1), slice(None, -2)
        s4 = slice(None, None), slice(1, -1), slice(1, -1)
        s5 = slice(None, None), slice(1, -1), slice(2, None)
        s6 = slice(None, None), slice(2, None), slice(None, -2)
        s7 = slice(None, None), slice(2, None), slice(1, -1)
        s8 = slice(None, None), slice(2, None), slice(2, None)

        # angle calculation
        y = np.empty(array.shape, dtype="f4")
        y[s4] = (
            array[s0]
            + 2 * array[s1]
            + array[s2]
            - array[s6]
            - 2 * array[s7]
            - array[s8]
        ) / yres

        x = np.empty(array.shape, dtype="f4")
        x[s4] = (
            array[s0]
            + 2 * array[s3]
            + array[s6]
            - array[s2]
            - 2 * array[s5]
            - array[s8]
        ) / xres

        with np.errstate(all="ignore"):
            xx_plus_yy = x * x + y * y
            aspect = np.arctan2(y, x)

            # shading
            cang = (
                math.sin(alt)
                - math.cos(alt) * zsf * np.sqrt(xx_plus_yy) * np.sin(aspect - az)
            ) / np.sqrt(1 + square_zsf * xx_plus_yy)

        cang = cang[..., 1:-1, 1:-1]
        result = np.where(cang <= 0, 0, 255 * cang).astype("u1")
        return {"values": result, "no_data_value": 256}

    def get_sources_and_requests(self, **request):
        new_request = expand_request_pixels(request, radius=1)
        if new_request is None:  # not an expandable request: do nothing
            return [(self.store, request)]

        # determine resolution
        bbox = request["bbox"]
        resolution = (
            (bbox[2] - bbox[0]) / request["width"],
            (bbox[3] - bbox[1]) / request["height"],
        )

        process_kwargs = dict(
            resolution=resolution,
            altitude=self.altitude,
            azimuth=self.azimuth,
            fill=self.fill,
        )

        return [(self.store, new_request), (process_kwargs, None)]


class Place(BaseSingle):
    """Place an input raster at given coordinates

    Note that if the store's projection is different from the requested one,
    the data will be reprojected before placing it at a different position.

    Args:
      store (RasterBlock): Raster that will be placed.
      place_projection (str): The projection in which this operation is done.
        This also specifies the projection of the ``anchor`` and
        ``coordinates`` args.
      anchor (list of 2 numbers): The anchor into the source raster that will
        be placed at given coordinates.
      coordinates (list of lists of 2 numbers): The target coordinates. The
        center of the bbox will be placed on each of these coordinates.
      statistic (str): What method to use to merge overlapping rasters. One of:
        {"last", "first", "count", "sum", "mean", "min",
        "max", "argmin", "argmax", "product", "std", "var", "p<number>"}

    Returns:
      RasterBlock with the source raster placed

    """

    def __init__(self, store, place_projection, anchor, coordinates, statistic="last"):
        if not isinstance(store, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(store)))
        try:
            get_sr(place_projection)
        except RuntimeError:
            raise ValueError(
                "'{}' is not a valid projection string".format(place_projection)
            )
        anchor = list(anchor)
        if len(anchor) != 2:
            raise ValueError("Expected 2 numbers in the 'anchor' parameter")
        for x in anchor:
            if not isinstance(x, (int, float)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        if coordinates is None or len(coordinates) == 0:
            coordinates = []
        else:
            coordinates = np.asarray(coordinates, dtype=float)
            if coordinates.ndim != 2 or coordinates.shape[1] != 2:
                raise ValueError(
                    "Expected a list of lists of 2 numbers in the "
                    "'coordinates' parameter"
                )
            coordinates = coordinates.tolist()
        check_statistic(statistic)
        super().__init__(store, place_projection, anchor, coordinates, statistic)

    @property
    def place_projection(self):
        return self.args[1]

    @property
    def anchor(self):
        return self.args[2]

    @property
    def coordinates(self):
        return self.args[3]

    @property
    def statistic(self):
        return self.args[4]

    @property
    def projection(self):
        """The native projection of this block.

        Only returns something if the place projection equals the store
        projection"""
        store_projection = self.store.projection
        if store_projection is None:
            return
        if get_sr(self.place_projection).IsSame(get_sr(store_projection)):
            return store_projection

    @property
    def geo_transform(self):
        """The native geo_transform of this block

        Returns None if the store projection and place projections differ."""
        if self.projection is not None:
            return self.store.geo_transform

    @property
    def extent(self):
        geometry = self.geometry
        if geometry is None:
            return
        if not geometry.GetSpatialReference().IsSame(EPSG4326):
            geometry = geometry.Clone()
            geometry.TransformTo(EPSG4326)
        x1, x2, y1, y2 = geometry.GetEnvelope()
        return x1, y1, x2, y2

    @property
    def geometry(self):
        """Combined geometry in this block's native projection. """
        store_geometry = self.store.geometry
        if store_geometry is None:
            return
        sr = get_sr(self.place_projection)
        if not store_geometry.GetSpatialReference().IsSame(sr):
            store_geometry = store_geometry.Clone()
            store_geometry.TransformTo(sr)
        _x1, _x2, _y1, _y2 = store_geometry.GetEnvelope()
        p, q = self.anchor
        P, Q = zip(*self.coordinates)
        x1, x2 = _x1 + min(P) - p, _x2 + max(P) - p
        y1, y2 = _y1 + min(Q) - q, _y2 + max(Q) - q
        return ogr.CreateGeometryFromWkt(POLYGON.format(x1, y1, x2, y2), sr)

    def get_sources_and_requests(self, **request):
        if request["mode"] != "vals":
            return ({"mode": request["mode"]}, None), (self.store, request)

        # transform the anchor and coordinates into the requested projection
        anchor = shapely_transform(
            Point(self.anchor), self.place_projection, request["projection"]
        ).coords[0]
        coordinates = [
            shapely_transform(
                Point(coord), self.place_projection, request["projection"]
            ).coords[0]
            for coord in self.coordinates
        ]

        # transform the source's extent
        extent_geometry = self.store.geometry
        if extent_geometry is None:
            # no geometry means: no data
            return (({"mode": "null"}, None),)
        sr = get_sr(request["projection"])
        if not extent_geometry.GetSpatialReference().IsSame(sr):
            extent_geometry = extent_geometry.Clone()
            extent_geometry.TransformTo(sr)
        xmin, xmax, ymin, ymax = extent_geometry.GetEnvelope()

        # compute the requested cellsize
        x1, y1, x2, y2 = request["bbox"]
        size_x = (x2 - x1) / request["width"]
        size_y = (y2 - y1) / request["height"]

        # point requests: never request the full source extent
        if size_x > 0 and size_y > 0:
            # check what the full source extent would require
            full_height = math.ceil((ymax - ymin) / size_y)
            full_width = math.ceil((xmax - xmin) / size_x)
            if full_height * full_width <= request["width"] * request["height"]:
                _request = request.copy()
                _request["width"] = full_width
                _request["height"] = full_height
                _request["bbox"] = (
                    xmin,
                    ymin,
                    xmin + full_width * size_x,
                    ymin + full_height * size_y,
                )
                process_kwargs = {
                    "mode": "warp",
                    "anchor": anchor,
                    "coordinates": coordinates,
                    "src_bbox": _request["bbox"],
                    "dst_bbox": request["bbox"],
                    "cellsize": (size_x, size_y),
                    "statistic": self.statistic,
                }
                return [(process_kwargs, None), (self.store, _request)]

        # generate a new (backwards shifted) bbox for each coordinate
        sources_and_requests = []
        filtered_coordinates = []
        for _x, _y in coordinates:
            bbox = [
                x1 + anchor[0] - _x,
                y1 + anchor[1] - _y,
                x2 + anchor[0] - _x,
                y2 + anchor[1] - _y,
            ]
            # check the overlap with the source's extent
            # Note that raster cells are defined [xmin, xmax) and (ymin, ymax]
            # so points precisely at xmax or ymin certainly do not have data.
            if bbox[0] >= xmax or bbox[1] > ymax or bbox[2] < xmin or bbox[3] <= ymin:
                continue
            filtered_coordinates.append((_x, _y))
            _request = request.copy()
            _request["bbox"] = bbox
            sources_and_requests.append((self.store, _request))
        if len(sources_and_requests) == 0:
            # No coordinates inside: we still need to return an array
            # of the correct shape. Send a time request to get the depth.
            _request = request.copy()
            _request["mode"] = "time"
            process_kwargs = {
                "mode": "empty",
                "dtype": self.dtype,
                "fillvalue": self.fillvalue,
                "width": request["width"],
                "height": request["height"],
                "statistic": self.statistic,
            }
            return [(process_kwargs, None), (self.store, _request)]
        process_kwargs = {"mode": "group", "statistic": self.statistic}
        return [(process_kwargs, None)] + sources_and_requests

    @staticmethod
    def process(process_kwargs, *multi):
        if process_kwargs["mode"] in {"meta", "time"}:
            return multi[0]
        if process_kwargs["mode"] == "null":
            return
        if process_kwargs["mode"] == "empty":
            data = multi[0]
            if data is None:
                return
            out_shape = (
                len(data["time"]),
                process_kwargs["height"],
                process_kwargs["width"],
            )
            out_no_data_value = process_kwargs["fillvalue"]
            out_dtype = process_kwargs["dtype"]
            stack = []
        elif process_kwargs["mode"] == "group":
            # We have a bunch of arrays that are already shifted. Stack them.
            stack = [data for data in multi if data is not None]
            if len(stack) == 0:
                return  # instead of returning nodata (because inputs are None)
        elif process_kwargs["mode"] == "warp":
            # There is a single 'source' raster that we are going to shift
            # multiple times into the result. The cellsize is already correct.
            data = multi[0]
            if data is None:
                return
            out_no_data_value = data["no_data_value"]
            source = data["values"]
            out_dtype = source.dtype

            # convert the anchor to pixels (indices inside 'source')
            anchor = process_kwargs["anchor"]
            src_bbox = process_kwargs["src_bbox"]
            size_x, size_y = process_kwargs["cellsize"]
            anchor_px = (
                (anchor[0] - src_bbox[0]) / size_x,
                (anchor[1] - src_bbox[1]) / size_y,
            )

            # compute the output shape
            x1, y1, x2, y2 = process_kwargs["dst_bbox"]
            coordinates = process_kwargs["coordinates"]
            dst_h = round((y2 - y1) / size_y)
            dst_w = round((x2 - x1) / size_x)
            src_d, src_h, src_w = source.shape
            out_shape = (src_d, dst_h, dst_w)

            # determine what indices in 'source' have data
            k, j, i = np.where(get_index(source, out_no_data_value))

            # place the data on each coordinate
            stack = []
            for x, y in coordinates:
                if i.size == 0:  # shortcut: no data at all to place
                    break
                # transform coordinate into pixels (indices in 'values')
                coord_px = (x - x1) / size_x, (y - y1) / size_y
                di = round(coord_px[0] - anchor_px[0])
                dj = round(coord_px[1] - anchor_px[1])
                # because of the y-axis inversion: dj is measured from the
                # other side of the array. if you draw it, you'll arrive at:
                dj = dst_h - src_h - dj

                if di <= -src_w or di >= dst_w or dj <= -src_h or dj >= dst_h:
                    # skip as it would shift completely outside
                    continue
                elif 0 <= di <= (dst_w - src_w) and 0 <= dj <= (dst_h - src_h):
                    # complete place
                    values = np.full(out_shape, out_no_data_value, out_dtype)
                    values[k, j + dj, i + di] = source[k, j, i]
                    stack.append({"values": values, "no_data_value": out_no_data_value})
                else:
                    # partial place
                    i_s = i + di
                    j_s = j + dj
                    m = (i_s >= 0) & (j_s >= 0) & (i_s < dst_w) & (j_s < dst_h)
                    if not m.any():
                        continue
                    values = np.full(out_shape, out_no_data_value, out_dtype)
                    values[k[m], j_s[m], i_s[m]] = source[k[m], j[m], i[m]]
                    stack.append({"values": values, "no_data_value": out_no_data_value})

        # merge the values_stack
        if len(stack) == 0:
            return {
                "values": np.full(out_shape, out_no_data_value, out_dtype),
                "no_data_value": out_no_data_value,
            }
        else:
            return reduce_rasters(stack, process_kwargs["statistic"])



"""
Module containing miscellaneous raster blocks.
"""
from osgeo import ogr
import numpy as np
from geopandas import GeoSeries

from shapely.geometry import box
from shapely.geometry import Point
from shapely.errors import WKTReadingError
from shapely.wkt import loads as load_wkt

from dask import config
from dask_geomodeling.geometry import GeometryBlock
from dask_geomodeling import utils

from .base import RasterBlock, BaseSingle


__all__ = [
    "Clip",
    "Classify",
    "Reclassify",
    "Mask",
    "MaskBelow",
    "Step",
    "Rasterize",
    "RasterizeWKT",
]


class Clip(BaseSingle):
    """
    Clip one raster to the extent of another raster.

    Takes two raster inputs, one raster ('store') whose values are returned in
    the output and one raster ('source') that is used as the extent. Cells of
    the 'store' raster are replaced with 'no data' if there is no data in the
    'source' raster.

    If the 'source' raster is a boolean raster, False will result in 'no data'.

    Note that the input rasters are required to have the same time resolution.

    Args:
      store (RasterBlock): Raster whose values are clipped
      source (RasterBlock): Raster that is used as the clipping mask

    Returns:
      RasterBlock with clipped values.
    """

    def __init__(self, store, source):
        if not isinstance(source, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(store)))
        # timedeltas are required to be equal
        if store.timedelta != source.timedelta:
            raise ValueError(
                "Time resolution of the clipping mask does not match that of "
                "the values raster. Consider using Snap."
            )
        super(Clip, self).__init__(store, source)

    @property
    def source(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        start = request.get("start", None)
        stop = request.get("stop", None)

        if start is not None and stop is not None:
            # limit request to self.period so that resulting data is aligned
            period = self.period
            if period is not None:
                request["start"] = max(start, period[0])
                request["stop"] = min(stop, period[1])

        return ((source, request) for source in self.args)

    @staticmethod
    def process(data, source_data):
        """ Mask store_data where source_data has no data """
        if data is None:
            return None

        if "values" not in data:
            return data

        # check if values contain data
        if np.all(data["values"] == data["no_data_value"]):
            return data

        # make the boolean mask
        if source_data is None:
            return None

        if source_data["values"].dtype == np.dtype("bool"):
            mask = ~source_data["values"]
        else:
            mask = source_data["values"] == source_data["no_data_value"]

        # adjust values
        values = data["values"].copy()
        values[mask] = data["no_data_value"]
        return {"values": values, "no_data_value": data["no_data_value"]}

    @property
    def extent(self):
        """Intersection of bounding boxes of 'store' and 'source'. """
        result, mask = [s.extent for s in self.args]
        if result is None or mask is None:
            return

        # return the overlapping box
        x1 = max(result[0], mask[0])
        y1 = max(result[1], mask[1])
        x2 = min(result[2], mask[2])
        y2 = min(result[3], mask[3])
        if x2 <= x1 or y2 <= y1:
            return None  # no overlap
        else:
            return x1, y1, x2, y2

    @property
    def geometry(self):
        """Intersection of geometries of 'store' and 'source'. """
        result, mask = [x.geometry for x in self.args]
        if result is None or mask is None:
            return
        sr = result.GetSpatialReference()
        if not mask.GetSpatialReference().IsSame(sr):
            mask = mask.Clone()
            mask.TransformTo(sr)
        result = result.Intersection(mask)
        if result.GetArea() == 0.0:
            return
        return result

    @property
    def period(self):
        """ Return period datetime tuple. """
        periods = [x.period for x in self.args]
        if any(period is None for period in periods):
            return None  # None precedes

        # multiple periods: return the overlapping period
        start = max([p[0] for p in periods])
        stop = min([p[1] for p in periods])
        if stop < start:
            return None  # no overlap
        else:
            return start, stop


class Mask(BaseSingle):
    """
    Replace values in a raster with a single constant value. 'no data' values
    are preserved.

    Args:
      store (RasterBlock): The raster whose values are to be converted.
      value (number): The constant value to be given to 'data' values.

    Returns:
      RasterBlock containing a single value
    """

    def __init__(self, store, value):
        if not isinstance(value, (float, int)):
            raise TypeError("'{}' object is not allowed".format(type(value)))
        super(Mask, self).__init__(store, value)

    @property
    def value(self):
        return self.args[1]

    @property
    def fillvalue(self):
        return 1 if self.value == 0 else 0

    @property
    def dtype(self):
        return self._dtype_from_value(self.value)

    @staticmethod
    def _dtype_from_value(value):
        if isinstance(value, float):
            return np.dtype("float32")
        elif value >= 0:
            return utils.get_uint_dtype(value)
        else:
            return utils.get_int_dtype(value)

    @staticmethod
    def process(data, value):
        if data is None or "values" not in data:
            return data

        index = utils.get_index(
            values=data["values"], no_data_value=data["no_data_value"]
        )

        fillvalue = 1 if value == 0 else 0
        dtype = Mask._dtype_from_value(value)

        values = np.full_like(data["values"], fillvalue, dtype=dtype)
        values[index] = value
        return {"values": values, "no_data_value": fillvalue}


class MaskBelow(BaseSingle):
    """
    Converts raster cells below the supplied value to 'no data'.

    Raster cells with values greater than or equal to the supplied value are
    returned unchanged.

    Args:
      store (RasterBlock): The raster whose values are to be masked.
      value (number): The constant value below which values are masked.

    Returns:
      RasterBlock with cells below the input value converted to 'no data'.
    """

    def __init__(self, store, value):
        if not isinstance(value, (float, int)):
            raise TypeError("'{}' object is not allowed".format(type(value)))
        super(MaskBelow, self).__init__(store, value)

    @staticmethod
    def process(data, value):
        if data is None or "values" not in data:
            return data
        values, no_data_value = data["values"].copy(), data["no_data_value"]
        values[values < value] = no_data_value
        return {"values": values, "no_data_value": no_data_value}


class Step(BaseSingle):
    """
    Apply a step function to a raster.

    This operation classifies the elements of a raster into three categories:
    less than, equal to, and greater than a value.

    The step function is defined as follows, with x being the value of a raster
    cell:

    - 'left' if *x < value*

    - 'at' if *x == value*

    - 'right' if *x > value*

    Args:
      store (RasterBlock): The input raster
      left (number): Value given to cells lower than the input value,
        defaults to 0
      right (number): Value given to cells higher than the input value,
        defaults to 1
      value (number): The constant value which raster cells are compared to,
        defaults to 0
      at (number): Value given to cells equal to the input value, defaults to
        the average of left and right

    Returns:
      RasterBlock containing three values; left, right and at.

    """

    def __init__(self, store, left=0, right=1, value=0, at=None):
        at = (left + right) / 2 if at is None else at
        for x in left, right, value, at:
            if not isinstance(x, (float, int)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(Step, self).__init__(store, left, right, value, at)

    @property
    def left(self):
        return self.args[1]

    @property
    def right(self):
        return self.args[2]

    @property
    def value(self):
        return self.args[3]

    @property
    def at(self):
        return self.args[4]

    @staticmethod
    def process(data, left, right, location, at):
        if data is None or "values" not in data:
            return data

        values, no_data_value = data["values"].copy(), data["no_data_value"]

        # determine boolean index arrays
        mask = values == no_data_value
        left_index = values < location
        at_index = values == location
        right_index = values > location
        # perform mapping
        values[left_index] = left
        values[at_index] = at
        values[right_index] = right
        # put no data values back
        values[mask] = no_data_value

        return {"values": values, "no_data_value": no_data_value}


class Classify(BaseSingle):
    """
    Classify raster data into binned categories

    Takes a RasterBlock and classifies its values based on bins. The bins are
    supplied as a list of increasing bin edges.

    For each raster cell this operation returns the index of the bin to which
    the raster cell belongs. The lowest possible output cell value is 0, which
    means that the input value was lower than the lowest bin edge. The highest
    possible output value is equal to the number of supplied bin edges.

    Args:
      store (RasterBlock): The raster whose cell values are to be classified
      bins (list): An increasing list of bin edges
      right (boolean): Whether the intervals include the right or the left bin
        edge, defaults to False.

    Returns:
      RasterBlock with classified values

    """

    def __init__(self, store, bins, right=False):
        if not isinstance(store, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(store)))
        if not hasattr(bins, "__iter__"):
            raise TypeError("'{}' object is not allowed".format(type(bins)))
        bins_arr = np.asarray(bins)
        if bins_arr.ndim != 1:
            raise TypeError("'bins' should be one-dimensional")
        if not np.issubdtype(bins_arr.dtype, np.number):
            raise TypeError("'bins' should be numeric")
        bins_diff = np.diff(bins)
        if not np.all(bins_diff > 0) or np.all(bins_diff < 0):
            raise TypeError("'bins' should be monotonic")
        super(Classify, self).__init__(store, bins_arr.tolist(), right)

    @property
    def bins(self):
        return self.args[1]

    @property
    def right(self):
        return self.args[2]

    @property
    def dtype(self):
        # with 254 bin edges, we have 255 bins, and we need 256 possible values
        # to include no_data
        return utils.get_uint_dtype(len(self.bins) + 2)

    @property
    def fillvalue(self):
        return utils.get_dtype_max(self.dtype)

    @staticmethod
    def process(data, bins, right):
        if data is None or "values" not in data:
            return data

        values = data["values"]
        dtype = utils.get_uint_dtype(len(bins) + 2)
        fillvalue = utils.get_dtype_max(dtype)

        result_values = np.digitize(values, bins, right).astype(dtype)
        result_values[values == data["no_data_value"]] = fillvalue

        return {"values": result_values, "no_data_value": fillvalue}


class Reclassify(BaseSingle):
    """
    Reclassify a raster of integer values.

    This operation can be used to reclassify a classified raster into desired
    values. Reclassification is done by supplying a list of [from, to] pairs.

    Args:
      store (RasterBlock): The raster whose cell values are to be reclassified
      bins (list): A list of [from, to] pairs defining the reclassification.
        The from values can be of bool or int datatype; the to values can be of
        int or float datatype
      select (boolean): Whether to set all non-reclassified cells to 'no data',
        defaults to False.

    Returns:
      RasterBlock with reclassified values
    """

    def __init__(self, store, data, select=False):
        dtype = store.dtype
        if dtype != bool and not np.issubdtype(dtype, np.integer):
            raise TypeError("The store must be of boolean or integer datatype")

        # validate "data"
        if not hasattr(data, "__iter__"):
            raise TypeError("'{}' object is not allowed".format(type(data)))
        try:
            source, target = self._data_as_ndarray(data)
        except ValueError:
            raise ValueError("Please supply a list of [from, to] values")
        # "from" can have bool or int dtype, "to" can also be float
        if source.dtype != bool and not np.issubdtype(source.dtype, np.integer):
            raise TypeError(
                "Cannot reclassify from value with type '{}'".format(source.dtype)
            )
        if len(np.unique(source)) != len(source):
            raise ValueError("There are duplicates in the reclassify values")
        if not np.issubdtype(target.dtype, np.number):
            raise TypeError(
                "Cannot reclassify to value with type '{}'".format(target.dtype)
            )
        # put 'data' into a list with consistent dtypes
        data = [list(x) for x in zip(source.tolist(), target.tolist())]

        if select is not True and select is not False:
            raise TypeError("'{}' object is not allowed".format(type(select)))
        super().__init__(store, data, select)

    @staticmethod
    def _data_as_ndarray(data):
        source, target = zip(*data)
        return np.asarray(source), np.asarray(target)

    @property
    def data(self):
        return self.args[1]

    @property
    def select(self):
        return self.args[2]

    @property
    def dtype(self):
        _, target = self._data_as_ndarray(self.data)
        return target.dtype

    @property
    def fillvalue(self):
        return utils.get_dtype_max(self.dtype)

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "dtype": self.dtype.str,
            "fillvalue": self.fillvalue,
            "data": self.data,
            "select": self.select,
        }
        return [(self.store, request), (process_kwargs, None)]

    @staticmethod
    def process(store_data, process_kwargs):
        if store_data is None or "values" not in store_data:
            return store_data

        no_data_value = store_data["no_data_value"]
        values = store_data["values"]
        source, target = Reclassify._data_as_ndarray(process_kwargs["data"])
        dtype = np.dtype(process_kwargs["dtype"])
        fillvalue = process_kwargs["fillvalue"]

        # add the nodata value to the source array and map it to the target
        # nodata
        if no_data_value is not None and no_data_value not in source:
            source = np.append(source, no_data_value)
            target = np.append(target, fillvalue)

        # sort the source and target values
        inds = np.argsort(source)
        source = source[inds]
        target = target[inds]

        # create the result array
        if process_kwargs["select"]:  # select = True: initialize with nodata
            result = np.full(values.shape, fillvalue, dtype=dtype)
        else:  # select = True: initialize with existing data
            result = values.astype(dtype)  # makes a copy

        # find all values in the source data that are to be mapped
        mask = np.in1d(values.ravel(), source)
        mask.shape = values.shape
        # place the target values (this also maps nodata values)
        result[mask] = target[np.searchsorted(source, values[mask])]
        return {"values": result, "no_data_value": fillvalue}


class Rasterize(RasterBlock):
    """
    Converts geometry source to raster

    This operation is used to transform GeometryBlocks into RasterBlocks. Here
    geometries (from for example a shapefile) are converted to a raster, using
    the values from one of the columns.

    Note that to rasterize floating point values, it is necessary to pass
    ``dtype="float"``.

    Args:
      source (GeometryBlock): The geometry source to be rasterized
      column_name (string): The name of the column whose values will be
        returned in the raster. If column_name is not provided, a boolean
        raster will be generated indicating where there are geometries.
      dtype (string): A numpy datatype specification to return the array.
        Defaults to 'int32' if column_name is provided, or to 'bool' otherwise.

    Returns:
      RasterBlock with values from 'column_name' or a boolean raster.

    See also:
      https://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html

    The global geometry-limit setting can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.geometry-limit": 100000})
    """

    def __init__(self, source, column_name=None, dtype=None, limit=None):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if column_name is not None and not isinstance(column_name, str):
            raise TypeError("'{}' object is not allowed".format(type(column_name)))
        if dtype is None:  # set default values
            dtype = "bool" if column_name is None else "int32"
        else:  # parse to numpy dtype and back to string
            dtype = str(np.dtype(dtype))
        if limit and not isinstance(limit, int):
            raise TypeError("'{}' object is not allowed".format(type(limit)))
        if limit and limit < 1:
            raise ValueError("Limit should be greater than 1")
        super(Rasterize, self).__init__(source, column_name, dtype, limit)

    @property
    def source(self):
        return self.args[0]

    @property
    def column_name(self):
        return self.args[1]

    @property
    def limit(self):
        return self.args[3]

    @property
    def dtype(self):
        return np.dtype(self.args[2])

    @property
    def fillvalue(self):
        return None if self.dtype == bool else utils.get_dtype_max(self.dtype)

    @property
    def period(self):
        return (self.DEFAULT_ORIGIN,) * 2

    @property
    def extent(self):
        return None

    @property
    def timedelta(self):
        return None

    @property
    def geometry(self):
        return None

    @property
    def projection(self):
        return None

    @property
    def geo_transform(self):
        return None

    def get_sources_and_requests(self, **request):
        # first handle the 'time' and 'meta' requests
        mode = request["mode"]
        if mode == "time":
            return [(self.period[-1], None), ({"mode": "time"}, None)]
        elif mode == "meta":
            return [(None, None), ({"mode": "meta"}, None)]
        elif mode != "vals":
            raise ValueError("Unknown mode '{}'".format(mode))

        # build the request to be sent to the geometry source
        x1, y1, x2, y2 = request["bbox"]
        width, height = request["width"], request["height"]

        # be strict about the bbox, it may lead to segfaults else
        if x2 == x1 and y2 == y1:  # point
            min_size = None
        elif x1 < x2 and y1 < y2:
            min_size = min((x2 - x1) / width, (y2 - y1) / height)
        else:
            raise ValueError("Invalid bbox ({})".format(request["bbox"]))

        limit = self.limit
        if self.limit is None:
            limit = config.get("geomodeling.geometry-limit")

        geom_request = {
            "mode": "intersects",
            "geometry": box(*request["bbox"]),
            "projection": request["projection"],
            "min_size": min_size,
            "limit": limit,
            "start": request.get("start"),
            "stop": request.get("stop"),
        }
        # keep some variables for use in process()
        process_kwargs = {
            "mode": "vals",
            "column_name": self.column_name,
            "dtype": self.dtype,
            "no_data_value": self.fillvalue,
            "width": width,
            "height": height,
            "bbox": request["bbox"],
        }
        return [(self.source, geom_request), (process_kwargs, None)]

    @staticmethod
    def process(data, process_kwargs):
        # first handle the time and meta requests
        mode = process_kwargs["mode"]
        if mode == "time":
            return {"time": [data]}
        elif mode == "meta":
            return {"meta": [None]}

        column_name = process_kwargs["column_name"]
        height = process_kwargs["height"]
        width = process_kwargs["width"]
        no_data_value = process_kwargs["no_data_value"]
        dtype = process_kwargs["dtype"]
        f = data["features"]

        # get the value column to rasterize
        if column_name is None:
            values = None
        else:
            try:
                values = f[column_name]
            except KeyError:
                if f.index.name == column_name:
                    values = f.index.to_series()
                else:
                    values = False

        if len(f) == 0 or values is False:  # there is no data to rasterize
            values = np.full((1, height, width), no_data_value, dtype=dtype)
            return {"values": values, "no_data_value": no_data_value}

        result = utils.rasterize_geoseries(
            geoseries=f["geometry"] if "geometry" in f else None,
            values=values,
            bbox=process_kwargs["bbox"],
            projection=data["projection"],
            height=height,
            width=width,
        )

        values = result["values"]

        # cast to the expected dtype if necessary
        cast_values = values.astype(process_kwargs["dtype"])

        # replace the nodata value if necessary
        if result["no_data_value"] != no_data_value:
            cast_values[values == result["no_data_value"]] = no_data_value

        return {"values": cast_values, "no_data_value": no_data_value}


class RasterizeWKT(RasterBlock):
    """Converts a single geometry to a raster mask

    Args:
      wkt (string): the WKT representation of a geometry
      projection (string): the projection of the geometry

    Returns:
      RasterBlock with True for cells that are inside the geometry.
    """

    def __init__(self, wkt, projection):
        if not isinstance(wkt, str):
            raise TypeError("'{}' object is not allowed".format(type(wkt)))
        if not isinstance(projection, str):
            raise TypeError("'{}' object is not allowed".format(type(projection)))
        try:
            load_wkt(wkt)
        except WKTReadingError:
            raise ValueError("The provided geometry is not a valid WKT")
        try:
            utils.get_sr(projection)
        except TypeError:
            raise ValueError("The provided projection is not a valid WKT")
        super().__init__(wkt, projection)

    @property
    def wkt(self):
        return self.args[0]

    @property
    def projection(self):
        return self.args[1]

    @property
    def dtype(self):
        return np.dtype("bool")

    @property
    def fillvalue(self):
        return None

    @property
    def period(self):
        return (self.DEFAULT_ORIGIN,) * 2

    @property
    def extent(self):
        return tuple(
            utils.shapely_transform(
                load_wkt(self.wkt), self.projection, "EPSG:4326"
            ).bounds
        )

    @property
    def timedelta(self):
        return None

    @property
    def geometry(self):
        return ogr.CreateGeometryFromWkt(self.wkt, utils.get_sr(self.projection))

    @property
    def geo_transform(self):
        return None

    def get_sources_and_requests(self, **request):
        # first handle the 'time' and 'meta' requests
        mode = request["mode"]
        if mode == "time":
            data = self.period[-1]
        elif mode == "meta":
            data = None
        elif mode == "vals":
            data = {"wkt": self.wkt, "projection": self.projection}
        else:
            raise ValueError("Unknown mode '{}'".format(mode))
        return [(data, None), (request, None)]

    @staticmethod
    def process(data, request):
        mode = request["mode"]
        if mode == "time":
            return {"time": [data]}
        elif mode == "meta":
            return {"meta": [None]}
        # load the geometry and transform it into the requested projection
        geometry = load_wkt(data["wkt"])
        if data["projection"] != request["projection"]:
            geometry = utils.shapely_transform(
                geometry, data["projection"], request["projection"]
            )

        # take a shortcut when the geometry does not intersect the bbox
        x1, y1, x2, y2 = request["bbox"]
        if (x1 == x2) and (y1 == y2):
            # Don't do box(x1, y1, x2, y2), this gives an invalid geometry.
            bbox_geom = Point(x1, y1)
        else:
            bbox_geom = box(x1, y1, x2, y2)
        if not geometry.intersects(bbox_geom):
            return {
                "values": np.full(
                    (1, request["height"], request["width"]), False, dtype=bool
                ),
                "no_data_value": None,
            }

        return utils.rasterize_geoseries(
            geoseries=GeoSeries([geometry]) if not geometry.is_empty else None,
            bbox=request["bbox"],
            projection=request["projection"],
            height=request["height"],
            width=request["width"],
        )



import glob
import os
import sys
import shutil
import logging
from contextlib import contextmanager

import json
import fiona
import geopandas
import tempfile
from dask.config import config
from dask.base import tokenize
from dask_geomodeling import utils

from .base import BaseSingle
from .parallelize import GeometryTiler

__all__ = ["GeometryFileSink", "to_file"]

logger = logging.getLogger(__name__)


def _to_json(value):
    if isinstance(value, (list, dict)):
        try:
            return json.dumps(value)
        except TypeError:
            return "<unable to export>"  # cannot output this value
    else:
        return value


class GeometryFileSink(BaseSingle):
    """Write geometry data to files in a specified directory

    Use GeometryFileSink.merge_files to merge tiles into one large file.

    Args:
      source (GeometryBlock): The block the data is coming from
      url (str): The target directory to put the files in. If relative, it is
        taken relative to the geomodeling.root setting.
      extension (str): The file extension (defines the format), one of
        ``{"shp", "gpkg", "geojson", "gml"}``. On some platforms, these options
        might be limited. For an accurate list, see
        ``GeometryFileSink.supported_extensions``.
      fields (dict): A mapping that relates column names to output file field
        names field names like ``{<output file field name>: <column name>}``.

    Relevant settings can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.root": '/my/output/data/path'})
    """

    supported_extensions = {
        k: v
        for k, v in (
            ("shp", "ESRI Shapefile"),
            ("gpkg", "GPKG"),
            ("geojson", "GeoJSON"),
            ("gml", "GML"),
        )
        # take only drivers which Fiona reports as writable
        if "w" in fiona.supported_drivers.get(v, "")
        # skip GPKG and GML on Win32 platform as it yields a segfault
        and not (sys.platform == "win32" and k in ("gpkg", "gml"))
    }

    def __init__(self, source, url, extension="shp", fields=None):
        safe_url = utils.safe_file_url(url)
        if not isinstance(extension, str):
            raise TypeError("'{}' object is not allowed".format(type(extension)))
        if len(extension) > 0 and extension[0] == ".":
            extension = extension[1:]  # chop off the dot
        if extension not in self.supported_extensions:
            raise ValueError("Format '{}' is unsupported".format(extension))
        if fields is None:
            fields = {x: x for x in source.columns if x != "geometry"}
        elif not isinstance(fields, dict):
            raise TypeError("'{}' object is not allowed".format(type(fields)))
        else:
            missing = set(fields.values()) - source.columns
            if missing:
                raise ValueError("Columns {} are not available".format(missing))
        super().__init__(source, safe_url, extension, fields)

    @property
    def url(self):
        return self.args[1]

    @property
    def extension(self):
        return self.args[2]

    @property
    def fields(self):
        return self.args[3]

    @property
    def columns(self):
        return {"saved"}

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "url": self.url,
            "fields": self.fields,
            "extension": self.extension,
            "hash": tokenize(request)[:7],
        }
        return [(self.source, request), (process_kwargs, None)]

    @staticmethod
    def process(data, process_kwargs):
        if "features" not in data or len(data["features"]) == 0:
            return data  # do nothing for non-feature or empty requests

        features = data["features"].copy()
        projection = data["projection"]
        path = utils.safe_abspath(process_kwargs["url"])
        fields = process_kwargs["fields"]
        extension = process_kwargs["extension"]
        driver = GeometryFileSink.supported_extensions[extension]

        # generate the directory if necessary
        os.makedirs(path, exist_ok=True)

        # the target file path is a deterministic hash of the request
        filename = ".".join([process_kwargs["hash"], extension])

        # add the index to the columns if necessary
        index_name = features.index.name
        if index_name in fields.values() and index_name not in features.columns:
            features[index_name] = features.index

        # copy the dataframe
        features = features[["geometry"] + list(fields.values())]

        # rename the columns
        features.columns = ["geometry"] + list(fields.keys())

        # serialize nested fields (lists or dicts)
        for col in fields.keys():
            series = features[col]
            if series.dtype == object or (
                str(series.dtype) == "category"
                and series.cat.categories.dtype == object
            ):
                features[col] = series.map(_to_json)

        # convert categoricals
        for col in fields.keys():
            series = features[col]
            if str(series.dtype) == "category":
                features[col] = series.astype(series.cat.categories.dtype)

        # GeoJSON needs reprojection to EPSG:4326
        if driver == "GeoJSON" and projection.upper() != "EPSG:4326":
            features = utils.geodataframe_transform(features, projection, "EPSG:4326")

        # generate the file
        features.to_file(os.path.join(path, filename), driver=driver)

        result = geopandas.GeoDataFrame(index=features.index)
        result["saved"] = True
        return {"features": result, "projection": projection}

    @staticmethod
    def merge_files(path, target, remove_source=False):
        """Merge files (the output of this Block) into one single file.

        Optionally removes the source files.
        """
        path = utils.safe_abspath(path)
        target = utils.safe_abspath(target)

        if os.path.exists(target):
            raise IOError("Target '{}' already exists".format(target))

        target_base, ext = os.path.splitext(target)
        source_paths = glob.glob(os.path.join(path, "*" + ext))
        if len(source_paths) == 0:
            raise IOError(
                "No source files found with matching extension '{}'".format(ext)
            )
        elif len(source_paths) == 1:
            # shortcut for single file. we need to copy/move all base_name.*
            # files (e.g. shapefiles have multiple files)
            source_base = os.path.splitext(source_paths[0])[0]
            move_or_copy = shutil.move if remove_source else shutil.copy
            for file_path in glob.glob(source_base + ".*"):
                move_or_copy(file_path, target_base + os.path.splitext(file_path)[1])
            return

        with utils.fiona_env():
            # first detect the driver etc
            with fiona.collection(source_paths[0], "r") as source:
                kwargs = {
                    "driver": source.driver,
                    "crs": source.crs,
                    "schema": source.schema,
                }
                if source.encoding:
                    kwargs["encoding"] = source.encoding

            with fiona.collection(target, "w", **kwargs) as out:
                for source_path in source_paths:
                    with fiona.collection(source_path, "r") as source:
                        out.writerecords(v for k, v in source.items())
                    if remove_source:
                        os.remove(source_path)

            if remove_source:
                try:
                    os.rmdir(path)
                except IOError:  # directory not empty: do nothing
                    pass


@contextmanager
def DryRunTempDir(*args, **kwargs):
    yield "/tmp/dummy"


def to_file(source, url, fields=None, tile_size=None, dry_run=False, **request):
    """Utility function to export data from a GeometryBlock to a file on disk.

    You need to specify the target file path as well as the extent geometry
    you want to save. Feature properties can be saved by providing a field
    mapping to the ``fields`` argument.

    To stay within memory constraints or to parallelize an operation, the
    ``tile_size`` argument can be provided.

    Args:
      source (GeometryBlock): the block the data is coming from
      url (str): The target file path. The extension determines the format. For
        supported formats, consult GeometryFileSink.supported_extensions.
      fields (dict): a mapping that relates column names to output file field
        names field names, ``{<output file field name>: <column name>, ...}``.
      tile_size (int): Optionally use this for large exports to stay within
        memory constraints. The export is split in tiles of given size (units
        are determined by the projection). Finally the tiles are merged.
      dry_run (bool): Do nothing, only validate the arguments.
      geometry (shapely Geometry): Limit exported objects to objects whose
        centroid intersects with this geometry.
      projection (str): The projection as a WKT string or EPSG code.
        Sets the projection of the geometry argument, the target
        projection of the data, and the tiling projection.
      mode (str): one of ``{"intersects", "centroid"}``, default "centroid"
      start (datetime): start date as UTC datetime
      stop (datetime): stop date as UTC datetime
      **request: see GeometryBlock request specification

    Relevant settings can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.root": '/my/output/data/path'})
      >>> config.set({"temporary_directory": '/my/alternative/tmp/dir'})
    """
    if "mode" not in request:
        request["mode"] = "centroid"

    path = utils.safe_abspath(url)
    extension = os.path.splitext(path)[1]

    TmpDir = DryRunTempDir if dry_run else tempfile.TemporaryDirectory
    with TmpDir(dir=config.get("temporary_directory", None)) as tmpdir:
        sink = GeometryFileSink(source, tmpdir, extension=extension, fields=fields)

        # wrap the sink in a GeometryTiler
        if tile_size is not None:
            sink = GeometryTiler(sink, tile_size, request["projection"])

        if dry_run:
            return

        # export the dataset to the tmpdir (full dataset or multiple tiles)
        sink.get_data(**request)

        # copy the file / gather the tiles to the target location
        GeometryFileSink.merge_files(tmpdir, path)



"""
Module providing a percentile function analogous to the functions provided by
scipy.ndimage.measurements.
"""

import numpy as np


def _safely_castable_to_int(dt):
    """Test whether the numpy data type `dt` can be safely cast to an int."""
    int_size = np.dtype(int).itemsize
    safe = (np.issubdtype(dt, np.signedinteger) and dt.itemsize <= int_size) or (
        np.issubdtype(dt, np.unsignedinteger) and dt.itemsize < int_size
    )
    return safe


def percentile(data, qval, labels=None, index=None):
    """
    Calculate a percentile of the array values over labeled regions.

    Parameters
    ----------
    input : array_like
        Array_like of values. For each region specified by `labels`, the
        percentile value of `input` over the region is computed.
    qval : float
        percentile to compute, which must be between 0 and 100 inclusive.
    labels : array_like, optional
        An array_like of integers marking different regions over which the
        percentile value of `input` is to be computed. `labels` must have the
        same shape as `input`. If `labels` is not specified, the percentile
        over the whole array is returned.
    index : array_like, optional
        A list of region labels that are taken into account for computing the
        percentiles. If index is None, the percentile over all elements where
        `labels` is non-zero is returned.

    Returns
    -------
    percentile : float or list of floats
        List of percentiles of `input` over the regions determined by `labels`
        and whose index is in `index`. If `index` or `labels` are not
        specified, a float is returned: the percentile value of `input` if
        `labels` is None, and the percentile value of elements where `labels`
        is greater than zero if `index` is None.

    Notes
    -----
    The function returns a Python list and not a Numpy array, use
    `np.array` to convert the list to an array.

    Examples
    --------
    >>> from scipy import ndimage
    >>> import numpy as np
    >>> a = np.array([[1, 2, 0, 1],
    ...               [5, 3, 0, 4],
    ...               [0, 0, 0, 7],
    ...               [9, 3, 0, 0]])
    >>> labels, labels_nb = ndimage.label(a)
    >>> labels
    array([[1, 1, 0, 2],
           [1, 1, 0, 2],
           [0, 0, 0, 2],
           [3, 3, 0, 0]], dtype=int32)
    >>> percentile(a, 75, labels=labels, index=np.arange(1, labels_nb + 1))
    [3.5, 5.5, 7.5]
    >>> percentile(a, 50)
    1.0
    >>> percentile(a, 50, labels=labels)
    3.0
    """
    data = np.asanyarray(data)

    def single_group(vals):
        return np.percentile(vals, qval)

    if labels is None:
        return single_group(data)

    # ensure input and labels match sizes
    data, labels = np.broadcast_arrays(data, labels)

    if index is None:
        mask = labels > 0
        return single_group(data[mask])

    if np.isscalar(index):
        mask = labels == index
        return single_group(data[mask])

    # remap labels to unique integers if necessary, or if the largest
    # label is larger than the number of values.
    if (
        not _safely_castable_to_int(labels.dtype)
        or labels.min() < 0
        or labels.max() > labels.size
    ):
        # remap labels, and indexes
        unique_labels, labels = np.unique(labels, return_inverse=True)
        idxs = np.searchsorted(unique_labels, index)

        # make all of idxs valid
        idxs[idxs >= unique_labels.size] = 0
        found = unique_labels[idxs] == index
    else:
        # labels are an integer type, and there aren't too many.
        idxs = np.asanyarray(index, int).copy()
        found = (idxs >= 0) & (idxs <= labels.max())

    idxs[~found] = labels.max() + 1

    # reorder data and labels, first by labels, then by data
    order = np.lexsort((data.ravel(), labels.ravel()))
    data = data.ravel()[order]
    labels = labels.ravel()[order]

    locs = np.arange(len(labels))
    lo = np.zeros(labels.max() + 2, int)
    lo[labels[::-1]] = locs[::-1]
    hi = np.zeros(labels.max() + 2, int)
    hi[labels] = locs
    lo = lo[idxs]
    hi = hi[idxs]
    # lo is an index to the lowest value in input for each label,
    # hi is an index to the largest value.

    # here starts the part that really diverts from scipy's median finder; the
    # linear interpolation method used corresponds to the default behaviour of
    # np.percentile().
    size = hi - lo + 1  # size of the group
    frac = (size - 1) * (qval / 100)  # fractional index relative to lo
    hi = lo - np.int64(-frac // 1)  # ceiled absolute index to data
    lo = lo + np.int64(frac // 1)  # floored absolute index to data
    part = frac % 1  # fractional part of index
    return (data[lo] + part * (data[hi] - data[lo])).tolist()



"""
Module containing blocks that parallelize raster blocks
"""
from itertools import product
import numpy as np

from .base import BaseSingle


__all__ = ["RasterTiler"]


class RasterTiler(BaseSingle):
    """Parallelize operations on a RasterBlock by tiling the request.

    Args:
      source (GeometryBlock): The source RasterBlock
      tile_size (int or list): The maximum size of a tile in pixels (cells)
         To specify different tile sizes for horizontal and vertical
         directions, provide the two as a list [width, height].

    Returns:
      RasterBlock
    """

    def __init__(self, source, tile_size):
        if hasattr(tile_size, "__iter__"):
            if len(tile_size) != 2:
                raise ValueError(
                    "'tile_size' should be a scalar or a list of length 2."
                )
            tile_size = [int(x) for x in tile_size]
        else:
            tile_size = [int(tile_size), int(tile_size)]
        if tile_size[0] <= 0 or tile_size[1] <= 0:
            raise ValueError("'tile_size' should be greater than 0")
        super().__init__(source, tile_size)

    @property
    def tile_size(self):
        return self.args[1]

    def get_sources_and_requests(self, **request):
        if request["mode"] != "vals":
            return [(None, None), (self.store, request)]

        # get the requested cell size
        x1, y1, x2, y2 = request["bbox"]
        cellsize_x = (x2 - x1) / request["width"]
        cellsize_y = (y2 - y1) / request["height"]
        if cellsize_x == 0 and cellsize_y == 0:
            # pass point requests through
            return [(None, None), (self.store, request)]

        # get tile size and compute tile edge coordinates
        tilesize_x = cellsize_x * self.tile_size[0]
        tilesize_y = cellsize_y * self.tile_size[1]
        x = np.arange(x1, x2, tilesize_x)
        y = np.arange(y1, y2, tilesize_y)

        # handle the 'leftover' tiles
        if x[-1] != x2:
            x = np.append(x, x2)
        if y[-1] != y2:
            y = np.append(y, y2)

        count_x, count_y = len(x) - 1, len(y) - 1

        # yield process_kwargs to be able piece back together the tiles later
        result = [
            (
                {
                    "dtype": self.dtype,
                    "fillvalue": self.fillvalue,
                    "shape_yx": (request["height"], request["width"]),
                    "count_xy": (count_x, count_y),
                    "tilesize_xy": self.tile_size,
                },
                None,
            )
        ]
        for i, j in product(range(count_x), range(count_y)):
            _request = {
                **request,
                "bbox": (x[i], y[j], x[i + 1], y[j + 1]),
                "width": int((x[i + 1] - x[i]) / cellsize_x),
                "height": int((y[j + 1] - y[j]) / cellsize_y),
            }
            result.append((self.store, _request))

        return result

    @staticmethod
    def process(process_kwargs, *all_data):
        if len(all_data) == 0:
            return
        elif process_kwargs is None:
            return all_data[0]  # for non-tiled / meta / time requests

        # go through all_data and get the temporal shape
        shape_yx = process_kwargs["shape_yx"]
        for data in all_data:
            if data is not None:
                shape = (data["values"].shape[0],) + shape_yx
                break
        else:
            return  # return None if all data is None

        # create the output array
        values = np.full(shape, process_kwargs["fillvalue"], process_kwargs["dtype"])

        # The tile order that was generated in the get_sources_and_request
        # starts at low x, low y and ends at high x, high y. As we are working
        # here with array indices [i, j] we need to take into account that the
        # vertical axis swaps direction: high y maps to low i.
        count_x, count_y = process_kwargs["count_xy"]
        tilesize_x, tilesize_y = process_kwargs["tilesize_xy"]
        for index, data in zip(product(range(count_x), range(count_y)), all_data):
            if data is None:
                continue
            vals = data["values"]
            j = index[0] * tilesize_x
            i = index[1] * tilesize_y
            values[:, -(i + vals.shape[1]) : -i or None, j : j + vals.shape[2]] = vals
        return {"values": values, "no_data_value": process_kwargs["fillvalue"]}



from osgeo import gdal, ogr, osr

gdal.UseExceptions()
ogr.UseExceptions()
osr.UseExceptions()

from . import config  # NOQA
from .core import Block, construct  # NOQA
from . import raster  # NOQA
from . import geometry  # NOQA



"""
Module containing merge operation that act on geometry blocks
"""
import pandas as pd
from shapely.geometry import box

from .base import GeometryBlock

__all__ = ["MergeGeometryBlocks"]


class MergeGeometryBlocks(GeometryBlock):
    """
    Merge two GeometryBlocks into one by index

    Provide two GeometryBlocks with the same original source to make sure they
    can be matched on index. The additional SeriesBlocks that have been added
    to the GeometryBlock will be combined to one GeometryBlock that contains
    all the information.

    Args:
      left (GeometryBlock): The left GeometryBlock to be combined.
      right (GeometryBlock): The right GeometryBlock to be combined.
      how (str, optional): The parameter that describes how the merge should
        be performed. There are four options:

        1. ``"left"``: The resulting GeometryBlock will have all the features
           that are present in the left GeometryBlock, no matter the features
           in the right GeometryBlock.
        2. ``"right"``: The resulting GeometryBlock will have all the features
           that are present in the right GeometryBlock, no matter the features
           in the left GeometryBlock.
        3. ``"inner"`` (default): The outcome will contain all the features
           that are present in both input GeometryBlocks. Features that are
           absent in one of the GeometryBlocks will be absent in the result.
        4. ``outer``: The result will contain all the features which are
           present in one of the input GeometryBlocks.

      suffixes (tuple, optional): Text to be added to the column
        names to distinguish whether they originate from the left or right
        GeometryBlock. Default: ``("", "_right")``.

    Returns:
      GeometryBlock that contains a combination of features and columns of the
      two input GeometryBlocks.
    """

    allow_how_joins = ("left", "right", "outer", "inner")

    def __init__(self, left, right, how="inner", suffixes=("", "_right")):
        if not isinstance(left, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(left)))
        if not isinstance(right, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(right)))
        if how not in self.allow_how_joins:
            raise KeyError(
                "'{}' is not part of the list of operations: "
                "{}".format(how, self.allow_how_joins)
            )
        if (
            not isinstance(suffixes[0], str)
            or not isinstance(suffixes[1], str)
            or len(suffixes) != 2
        ):
            raise TypeError("'{}' object is not " "allowed".format(type(suffixes)))
        super().__init__(left, right, how, suffixes)

    @property
    def left(self):
        return self.args[0]

    @property
    def right(self):
        return self.args[1]

    @property
    def how(self):
        return self.args[2]

    @property
    def suffixes(self):
        return self.args[3]

    @property
    def columns(self):
        left = self.left.columns
        right = self.right.columns
        result = left ^ right  # column in left or right, not both
        overlap = left & right
        for col in overlap:
            result |= {col + self.suffixes[0], col + self.suffixes[1]}
        return result

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "how": self.how,
            "suffixes": self.suffixes,
            "mode": request["mode"],
        }
        return [(self.left, request), (self.right, request), (process_kwargs, None)]

    @staticmethod
    def process(left, right, kwargs):
        mode = kwargs["mode"]
        how = kwargs["how"]
        projection = left["projection"]

        if mode == "intersects" or mode == "centroid":
            merged = pd.merge(
                left["features"],
                right["features"],
                how=kwargs.get("how"),
                suffixes=kwargs.get("suffixes"),
                left_index=True,  # we merge by index, left and right.
                right_index=True,
            )
            return {"features": merged, "projection": projection}
        elif mode == "extent":
            if how == "left":
                return left

            elif how == "right":
                return right

            elif how == "inner":
                values = None
                if left["extent"] and right["extent"]:
                    left_shape = box(*left["extent"])
                    right_shape = box(*right["extent"])
                    extent = left_shape.intersection(right_shape)
                    if not extent.is_empty:
                        values = extent.bounds
                return {"extent": values, "projection": projection}

            elif how == "outer":
                values = None
                if left["extent"] and right["extent"]:
                    left_shape = box(*left["extent"])
                    right_shape = box(*right["extent"])
                    values = left_shape.union(right_shape).bounds
                elif left["extent"]:
                    values = left["extent"]
                elif right["extent"]:
                    values = right["extent"]
                return {"extent": values, "projection": projection}



"""
Module containing geometry block operations that act on non-geometry fields
"""
import numpy as np
import pandas as pd
import operator
from .base import GeometryBlock, SeriesBlock, BaseSingleSeries

__all__ = [
    "Classify",
    "ClassifyFromColumns",
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "FloorDivide",
    "Power",
    "Modulo",
    "Equal",
    "NotEqual",
    "Greater",
    "GreaterEqual",
    "Less",
    "LessEqual",
    "And",
    "Or",
    "Xor",
    "Invert",
    "Where",
    "Mask",
    "Round",
]


class Classify(BaseSingleSeries):
    """
    Classify a value column into different bins

    For example: every value below 3 becomes "A", every value between 3 and 5
    becomes "B", and every value above 5 becomes "C".

    The provided SeriesBlock will be classified according to the given
    classification parameters. These parameters consist of two lists, one with
    the edges of the classification bins (i.e. ``[3, 5]``) and one with the
    desired class output (i.e. ``["A", "B", "C"]``). The input data is then
    compared to the classification bins. In this example a value 1 is below 3
    so it gets class ``"A"``. A value 4 is between 3 and 5 so it gets label
    ``"B"``.

    How values outside of the bins are classified depends on the length of the
    labels list. If the length of the labels equals the length of the binedges
    plus 1 (the above example), then values outside of the bins are classified
    to the first and last elements of the labels list. If the length of the
    labels equals the length of the bins minus 1, then values outside of the
    bins are classified to 'no data'.

    Args:
      source (SeriesBlock): The (numeric) data which should be classified.
      bins (list): The edges of the classification intervals
        (i.e. ``[3, 5]``).
      labels (list): The classification returned if a value falls in a specific
        bin (i.e. ``["A", "B", "C"]``). The length of this list is either one
        larger or one less than the length of the ``bins`` argument. Labels
        should be unique. If labels are numeric, they are always
        converted to float to be able to deal with NaN values.
      right (boolean, optional): Determines what side of the intervals are
        closed. Defaults to True (the right side of the bin is closed so a
        value assigned to the bin on the left if it is exactly on a bin edge).

    Returns:
      A SeriesBlock with classified values instead of the original numbers.
    """

    def __init__(self, source, bins, labels, right=True):
        if not isinstance(bins, list):
            raise TypeError("'{}' object is not allowed".format(type(bins)))
        if not isinstance(labels, list):
            raise TypeError("'{}' object is not allowed".format(type(labels)))
        if not isinstance(right, bool):
            raise TypeError("'{}' object is not allowed".format(type(right)))
        bins_arr = np.asarray(bins)
        if bins_arr.ndim != 1:
            raise TypeError("'bins' must be one-dimensional")
        if (np.diff(bins) < 0).any():
            raise ValueError("'bins' must increase monotonically.")
        if len(labels) not in (len(bins) - 1, len(bins) + 1):
            raise ValueError(
                "Expected {} or {} labels, got {}".format(
                    len(bins) - 1, len(bins) + 1, len(labels)
                )
            )
        if len(set(labels)) != len(labels):
            raise ValueError("Labels should be unique")
        super().__init__(source, bins, labels, right)

    @property
    def bins(self):
        return self.args[1]

    @property
    def labels(self):
        return self.args[2]

    @property
    def right(self):
        return self.args[3]

    @staticmethod
    def process(series, bins, labels, right):
        open_bounds = len(labels) == len(bins) + 1
        if open_bounds:
            bins = np.concatenate([[-np.inf], bins, [np.inf]])
        if series.dtype == object:
            series = series.fillna(value=np.nan)
        result = pd.cut(series, bins, right, labels)

        # Transform from categorical to whatever suits the "labels". The
        # dtype has to be able to accomodate NaN as well.
        result = result.astype(pd.Series(labels + [np.nan]).dtype)

        if open_bounds:
            # patch the result, we actually want to classify np.inf
            if right:
                result[series == -np.inf] = labels[0]
            else:
                result[series == np.inf] = labels[-1]
        return result


class ClassifyFromColumns(SeriesBlock):
    """
    Classify a continuous-valued geometry property based on bins located in
    other columns.

    See :class:``dask_geomodeling.geometry.field_operations.Classify`` for
    further information.

    Args:
      source (GeometryBlock):The GeometryBlock which contains the column which
        should be clasified as well as columns with the bin edges.
      value_column (str): The column with (float) data which should be
        classified.
      bin_columns (list): A list of columns that contain the bins for the
        classification. The order of the columns should be from low to high
        values.
      labels (list): The classification returned if a value falls in a specific
        bin (i.e. ``["A", "B", "C"]``). The length of this list is either one
        larger or one less than the length of the ``bins`` argument. Labels
        should be unique. If labels are numeric, they are always
        converted to float to be able to deal with NaN values.
      right (boolean, optional): Determines what side of the intervals are
        closed. Defaults to True (the right side of the bin is closed so a
        value assigned to the bin on the left if it is exactly on a bin edge).

    Returns:
      A SeriesBlock with classified values instead of the original floats.

    """

    def __init__(self, source, value_column, bin_columns, labels, right=True):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(value_column, str):
            raise TypeError("'{}' object is not allowed".format(type(value_column)))
        if not isinstance(bin_columns, list):
            raise TypeError("'{}' object is not allowed".format(type(bin_columns)))
        if not isinstance(labels, list):
            raise TypeError("'{}' object is not allowed".format(type(labels)))
        if not isinstance(right, bool):
            raise TypeError("'{}' object is not allowed".format(type(right)))
        missing_columns = (set(bin_columns) | {value_column}) - source.columns
        if missing_columns:
            raise KeyError("Columns '{}' are not present".format(missing_columns))
        if len(labels) not in (len(bin_columns) - 1, len(bin_columns) + 1):
            raise ValueError(
                "Expected {} or {} labels, got {}".format(
                    len(bin_columns) - 1, len(bin_columns) + 1, len(labels)
                )
            )
        if len(set(labels)) != len(labels):
            raise ValueError("Labels should be unique")
        super().__init__(source, value_column, bin_columns, labels, right)

    @property
    def source(self):
        return self.args[0]

    @property
    def value_column(self):
        return self.args[1]

    @property
    def bin_columns(self):
        return self.args[2]

    @property
    def labels(self):
        return self.args[3]

    @property
    def right(self):
        return self.args[4]

    @staticmethod
    def process(data, value_column, bin_columns, labels, right):
        if "features" not in data or len(data["features"]) == 0:
            return pd.Series([], dtype=float)
        features = data["features"]
        series = features[value_column]
        if series.dtype == object:
            series = series.fillna(value=np.nan)
        values = series.values
        bins = features[bin_columns].values
        n_bins = len(bin_columns)

        # Check in which bin every value is. because bins may be different for
        # each value, searchsorted is not an option. We assume that bins are
        # sorted in increasing order. Checking that would be costly.
        with np.errstate(invalid="ignore"):  # comparison to NaN is OK here
            if right:
                indices = np.sum(values[:, np.newaxis] > bins, axis=1)
            else:
                indices = np.sum(values[:, np.newaxis] >= bins, axis=1)

        # If values was NaN this is now assigned the value 0 (the first bin).
        # Convert to the last label so that we can map it later to NaN
        if len(labels) == n_bins + 1:
            indices[np.isnan(values)] = len(labels)
        else:
            # If we have e.g. 2 labels and 3 bins, the outside intervals are
            # closed. Therefore, indices 0 and 3 do not map to a bin. Index 0
            # also covers the values = NaN situation.
            indices -= 1  # indices become -1, 0, 1, 2
            indices[indices == -1] = len(labels)  # -1 --> 2

        # Convert indices to labels, append labels with with np.nan to cover
        # unclassified data.
        labeled_data = pd.Series(labels + [np.nan]).loc[indices]
        # Set the index to the features index
        labeled_data.index = features.index
        return labeled_data


class BaseFieldOperation(BaseSingleSeries):
    """Base block for basic operations between series from a GeometryBlock"""

    def __init__(self, source, other):
        if not isinstance(other, (SeriesBlock, int, float, bool)):
            raise TypeError("'{}' object is not allowed".format(type(other)))
        super().__init__(source, other)

    @property
    def other(self):
        return self.args[1]

    @staticmethod
    def process(source, other):
        raise NotImplementedError()  # implement this in subclasses


class Add(BaseFieldOperation):
    """
    Element-wise addition of SeriesBlock or number to another SeriesBlock.

    Args:
      source (SeriesBlock): First addition term
      other (SeriesBlock or number): Second addition term

    Returns:
      SeriesBlock
    """

    process = staticmethod(operator.add)


class Subtract(BaseFieldOperation):
    """
    Element-wise subtraction of SeriesBlock or number with another SeriesBlock.

    Note that if you want to subtract a SeriesBlock from a constant value (like
    ``4 - series``, you have to do ``Add(Multiply(series, -1), 4)``.

    Args:
      source (SeriesBlock): First subtraction term
      other (SeriesBlock or number): Second subtraction term

    Returns:
      SeriesBlock
    """

    process = staticmethod(operator.sub)


class Multiply(BaseFieldOperation):
    """
    Element-wise multiplication of SeriesBlock or number with another
    SeriesBlock.

    Args:
      source (SeriesBlock): First multiplication factor
      other (SeriesBlock or number): Second multiplication factor

    Returns:
      SeriesBlock
    """

    process = staticmethod(operator.mul)


class Divide(BaseFieldOperation):
    """
    Element-wise division of SeriesBlock or number with another SeriesBlock.

    Note that if you want to divide a constant value by a SeriesBlock (like
    ``3 / series``, you have to do ``Multiply(3, Power(series, -1))``.

    Args:
      source (SeriesBlock): Numerator
      other (SeriesBlock or number): Denominator

    Returns:
      SeriesBlock
    """

    process = staticmethod(operator.truediv)


class FloorDivide(BaseFieldOperation):
    """
    Element-wise integer division of SeriesBlock or number with another
    SeriesBlock.

    The outcome of the division is converted to the closest integer below (i.e.
    3.4 becomes 3, 3.9 becomes 3 and -3.4 becomes -4)

    Args:
      source (SeriesBlock): Numerator
      other (SeriesBlock or number): Denominator

    Returns:
      SeriesBlock
    """

    process = staticmethod(operator.floordiv)


class Power(BaseFieldOperation):
    """
    Element-wise raise a SeriesBlock to the power of a number or another
    SeriesBlock.

    For example, the inputs ``[2, 4]`` and ``2`` will give output ``[4, 16]``.

    Args:
      source (SeriesBlock): Base
      other (SeriesBlock or number): Exponent

    Returns:
      SeriesBlock
    """

    def __init__(self, source, other):
        # the float(other) will raise a TypeError if necessary
        super().__init__(source, float(other))

    process = staticmethod(operator.pow)


class Modulo(BaseFieldOperation):
    """
    Element-wise modulo (remainder after division) of SeriesBlock or number
    with another SeriesBlock.

    Example: if the input is ``[31, 5.3, -4]`` and the modulus is ``3``, the
    outcome would be ``[1, 2.3, 2]``. The outcome is always postive and less
    than the modulus.

    Args:
      source (SeriesBlock): Number
      other (SeriesBlock or number): Modulus

    Returns:
      SeriesBlock
    """

    process = staticmethod(operator.mod)


class Equal(BaseFieldOperation):
    """
    Determine whether a SeriesBlock and a second SeriesBlock or a constant
    value are equal.

    Note that 'no data' does not equal 'no data'.

    Args:
      source (SeriesBlock): First comparison term
      other (SeriesBlock or number): Second comparison term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.eq)


class NotEqual(BaseFieldOperation):
    """
    Determine whether a SeriesBlock and a second SeriesBlock or a constant
    value are not equal.

    Note that 'no data' does not equal 'no data'.

    Args:
      source (SeriesBlock): First comparison term
      other (SeriesBlock or number): Second comparison term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.ne)


class Greater(BaseFieldOperation):
    """
    Determine for each value in a SeriesBlock whether it is greater than a
    comparison value from a SeriesBlock or constant.

    Args:
      source (SeriesBlock): First comparison term
      other (SeriesBlock or number): Second comparison term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.gt)


class GreaterEqual(BaseFieldOperation):
    """
    Determine for each value in a SeriesBlock whether it is greater than or
    equal to a comparison value from a SeriesBlock or constant.

    Args:
      source (SeriesBlock): First comparison term
      other (SeriesBlock or number): Second comparison term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.ge)


class Less(BaseFieldOperation):
    """
    Determine for each value in a SeriesBlock whether it is less than a
    comparison value from a SeriesBlock or constant.

    Args:
      source (SeriesBlock): First comparison term
      other (SeriesBlock or number): Second comparison term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.lt)


class LessEqual(BaseFieldOperation):
    """
    Determine for each value in a SeriesBlock whether it is less than or equal
    to a comparison value from a SeriesBlock or constant.

    Args:
      source (SeriesBlock): First comparison term
      other (SeriesBlock or number): Second comparison term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.le)


class BaseLogicOperation(BaseFieldOperation):
    """Base block for logic operations between columns in a GeometryBlock"""

    def __init__(self, source, other):
        if not isinstance(other, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(other)))
        super().__init__(source, other)


class And(BaseLogicOperation):
    """
    Perform an elementwise logical AND between two SeriesBlocks.

    If a feature has a True value in both SeriesBlocks, True is returned, else
    False is returned.

    Args:
      source (SeriesBlock): First boolean term
      other (SeriesBlock): Second boolean term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.and_)


class Or(BaseLogicOperation):
    """
    Perform an elementwise logical OR between two SeriesBlocks.

    If a feature has a True value in any of the input SeriesBlocks, True is
    returned, else False is returned.

    Args:
      source (SeriesBlock): First boolean term
      other (SeriesBlock): Second boolean term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.or_)


class Xor(BaseLogicOperation):
    """
    Perform an elementwise logical exclusive OR between two SeriesBlocks.

    If a feature has a True value in precisely one of the input SeriesBlocks,
    True is returned, else False is returned.

    Args:
      source (SeriesBlock): First boolean term
      other (SeriesBlock): Second boolean term

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.xor)


class Invert(BaseSingleSeries):
    """
    Invert a boolean SeriesBlock (swap True and False)

    Args:
      source (SeriesBlock): SeriesBlock with boolean values.

    Returns:
      SeriesBlock with boolean values
    """

    process = staticmethod(operator.inv)


class Where(BaseSingleSeries):
    """
    Replace values in a SeriesBlock where values in another SeriesBlock are
    False.

    Provide a source SeriesBlock, a conditional SeriesBlock (True/False) and a
    replacement value which can either be a SeriesBlock or a constant value.
    All entries in the source that correspond to a True value in the
    conditional are left unchanged. The values in the source that correspond to
    a False value in the conditional are replaced with the value from 'other'.

    Args:
      source (SeriesBlock): Source SeriesBlock that is going to be updated
      cond (SeriesBlock): Conditional SeriesBlock that determines
        whether features in the source SeriesBlock will be updated. If this
        is not boolean (True/False), then all data values (including 0) are
        interpreted as True. Missing values are always interpeted as False.
      other (SeriesBlock or constant): The value that should be used as a
        replacement for the source SeriesBlock where the conditional
        SeriesBlock is False.

    Returns:
      SeriesBlock with updated values where condition is False.
    """

    def __init__(self, source, cond, other):
        if not isinstance(cond, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(cond)))
        super().__init__(source, cond, other)

    @property
    def cond(self):
        return self.args[1]

    @property
    def other(self):
        return self.args[2]

    @staticmethod
    def process(source, cond, other):
        if cond.dtype != bool:
            cond = ~pd.isnull(cond)
        return source.where(cond, other)


class Mask(BaseSingleSeries):
    """
    Replace values in a SeriesBlock where values in another SeriesBlock are
    True.

    Provide a source SeriesBlock, a conditional SeriesBlock (True/False) and a
    replacement value which can either be a SeriesBlock or a constant value.
    All entries in the source that correspond to a True value in the
    conditional are left unchanged. The values in the source that correspond to
    a True value in the conditional are replaced with the value from 'other'.

    Args:
      source (SeriesBlock): Source SeriesBlock that is going to be updated
      cond (SeriesBlock): Conditional SeriesBlock that determines
        whether features in the source SeriesBlock will be updated. If this
        is not boolean (True/False), then all data values (including 0) are
        interpreted as True. Missing values are always interpeted as False.
      other (SeriesBlock or constant): The value that should be used as a
        replacement for the source SeriesBlock where the conditional
        SeriesBlock is True.

    Returns:
      SeriesBlock with updated values where condition is True.
    """

    def __init__(self, source, cond, other):
        if not isinstance(cond, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(cond)))
        super().__init__(source, cond, other)

    @property
    def cond(self):
        return self.args[1]

    @property
    def other(self):
        return self.args[2]

    @staticmethod
    def process(source, cond, other):
        if cond.dtype != bool:
            cond = ~pd.isnull(cond)
        return source.mask(cond, other)


class Round(BaseSingleSeries):
    """
    Round each value in a SeriesBlock to the given number of decimals

    Args:
      source (SeriesBlock): SeriesBlock with float data that is rounded to the
        provided number of decimals.
      decimals (int, optional): number of decimal places to round to
        (default: 0). If decimals is negative, it specifies the number of
        positions to the left of the decimal point.

    Returns:
      SeriesBlock with rounded values.
    """

    def __init__(self, source, decimals=0):
        if not isinstance(decimals, int):
            raise TypeError("'{}' object is not allowed".format(type(decimals)))
        super().__init__(source, decimals)

    process = staticmethod(np.around)


class Interp(BaseSingleSeries):
    """One-dimensional piecewise linear interpolation.

    Given a list of datapoints corresponding to x and f(x), compute the
    one-dimensional piecewise linear interpolant f(x) for each input x
    from ``source``.

    Args:
      source (SeriesBlock): Source data (x-coordinates) to interpolate
      xp (list): The x-coordinates of the data points, must be increasing.
      fp (list): The y-coordinates of the data points, same length as ``xp``.
      left (number, optional): Value to return when an x-coordinate from
        ``source`` is below than the first ``xp``. Defaults to the first
        ``fp``.
      right (number, optional): Value to return when an x-coordinate from
        ``source`` is greater than the last ``xp``. Defaults to the last
        ``fp``.
    """

    def __init__(self, source, xp, fp, left=None, right=None):
        xp = [float(x) for x in xp]
        fp = [float(x) for x in fp]
        if left is not None:
            left = float(left)
        if right is not None:
            right = float(right)
        if np.any(np.diff(xp) < 0):
            raise ValueError("xp must be monotonically increasing")
        super().__init__(source, xp, fp, left, right)

    @staticmethod
    def process(data, xp, fp, left, right):
        result = np.interp(data, xp, fp, left, right)
        return pd.Series(result, index=data.index)


class Choose(BaseSingleSeries):
    """Construct a SeriesBlock by choosing values from multiple SeriesBlocks.

    For example, consider we have created ``Choose(X, A, B, C)``. The four
    input SeriesBlock contain the following values:

    - X: ``[0, 2, 1]``
    - A: ``[1, 2, 3]``
    - B: ``[4, 5, 6]``
    - C: ``[7, 8, 9]``

    The result will be ``[1, 8, 6]``: the first entry has X=0, so A is
    selected whose first entry is ``1``. The second entry has X=2, so B
    is selected, the third will select C, etc. Note that the choice series
    A, B and C can be of any type.

    If any value in ``source`` is out of bounds (below 0 or larger than the
    number of arguments), 'no data' will be filled in.

    Args:
      source (SeriesBlock): SeriesBlock having integers from 0 to n - 1 with n
        being the number of choices.
      *choices (SeriesBlock): Multiple series containing the values to choose

    Returns:
       SeriesBlock with values from ``choices``
    """

    def __init__(self, source, *choices):
        if not len(choices) >= 2:
            raise ValueError("The number of choices must be greater than one.")
        if not all([isinstance(choice, SeriesBlock) for choice in choices]):
            raise TypeError("All choices must be SeriesBlock objects")
        super().__init__(source, *choices)

    @property
    def choices(self):
        return self.args[1:]

    @staticmethod
    def process(source, *choices):
        result = pd.Series(np.nan, index=source.index)
        for i, choice in enumerate(choices):
            mask = source == i
            result[mask] = choice[source.index[mask]]
        return result



"""
Module containing geometry block constructive operations
"""
import numbers

from dask_geomodeling.utils import Extent, shapely_transform, transform_extent

from .base import BaseSingle

__all__ = ["Buffer", "Simplify"]


class Buffer(BaseSingle):
    """
    Buffer ('expand') geometries with a given value.

    A GeometryBlock and a buffer distance are provided. Each feature in the
    GeometryBlock is buffered with the distance provided, resulting in updated
    geometries.

    Args:
      source (GeometryBlock): The source GeometryBlock whose geometry will be
        updated.
      distance (float): The distance used to buffer all features. The distance
        is measured in the unit of the given projection (e.g. m, °).
      projection (str): The projection used in the operation provided in the
        format: ``"EPSG:28992"``.
      resolution (integer, optional): The resolution of the buffer provided as
        the number of points used to represent a quarter of a circle. The
        default value is ``16``.

    Returns:
      GeometryBlock with buffered geometries.

    """

    def __init__(self, source, distance, projection, resolution=16):
        if not isinstance(distance, numbers.Real):
            raise TypeError("Argument 'distance' must be a float or int.")
        if not isinstance(projection, str):
            raise TypeError("Argument 'projection' must be a str.")
        if not isinstance(resolution, int):
            raise TypeError("Argument 'resolution' must be an int.")
        super().__init__(source, distance, projection, resolution)

    @property
    def distance(self):
        return self.args[1]

    @property
    def projection(self):
        return self.args[2]

    @property
    def resolution(self):
        return self.args[3]

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "distance": self.distance,
            "buf_srs": self.projection,
            "resolution": self.resolution,
        }
        return [(self.source, request), (process_kwargs, None)]

    @staticmethod
    def process(data, kwargs):
        if "features" in data:
            if len(data["features"]) == 0:
                return data
            req_srs = data["projection"]
            buf_srs = kwargs["buf_srs"]
            features = data["features"].set_geometry(
                data["features"]
                .geometry.apply(shapely_transform, args=(req_srs, buf_srs))
                .buffer(distance=kwargs["distance"], resolution=kwargs["resolution"])
                .apply(shapely_transform, args=(buf_srs, req_srs))
            )
            return {"features": features, "projection": req_srs}
        elif "extent" in data:
            if not data["extent"]:
                return data
            req_srs = data["projection"]
            buf_srs = kwargs["buf_srs"]
            distance = kwargs["distance"]
            extent = transform_extent(data["extent"], req_srs, buf_srs)
            extent = Extent(extent, buf_srs).buffered(distance).bbox
            extent = transform_extent(extent, buf_srs, req_srs)
            return {"extent": extent, "projection": req_srs}
        else:
            raise NotImplementedError("Dunno this mode!")


class Simplify(BaseSingle):
    """
    Simplify geometries, mainly to make them computationally more efficient.

    Provide a GeometryBlock and a tolerance value to simplify the geometries.
    As a result all features in the GeometryBlock are simplified.

    Args:
      source (GeometryBlock): Source of the geometries to be simplified.
      tolerance (float): The tolerance used in the simplification. If no
        tolerance is given the ``"min_size"`` request parameter is used.
      preserve_topology (boolean, optional): Determines whether the topology
        should be preserved in the operation. Defaults to ``True``.

    Returns:
      GeometryBlock which was provided as input with a simplified geometry.

    """

    def __init__(self, source, tolerance=None, preserve_topology=True):
        if tolerance is not None:
            tolerance = float(tolerance)
        super().__init__(source, tolerance, bool(preserve_topology))

    @property
    def tolerance(self):
        return self.args[1]

    @property
    def preserve_topology(self):
        return self.args[2]

    def get_sources_and_requests(self, **request):
        process_kwargs = {
            "tolerance": self.tolerance or request.get("min_size") or 0.0,
            "preserve_topology": self.preserve_topology,
        }
        return [(self.source, request), (process_kwargs, None)]

    @staticmethod
    def process(data, kwargs):
        if "features" not in data:
            # basically assumes that the extent will not change during simplify
            return data
        features = data["features"].set_geometry(
            data["features"].geometry.simplify(**kwargs)
        )  # for 'intersects' and 'centroid' mode.
        return {"features": features, "projection": data["projection"]}
